"""
Weird one-off functions that don't have a good home.
"""
import xarray
import numpy as np

def load_masks_from_mds(grid_dir):
    import xmitgcm
    ds = xmitgcm.open_mdsdataset(grid_dir, iters=None, geometry='llc')

    points = ['W', 'W', 'S']
    masks = [ds['hfac' + point].reset_coords(drop=True).rename('mask' + point)
             for point in points]
    ds_mask = xr.merge(masks)
    for c in ds_mask.coords:
        ds_mask[c] = ds_mask[c].astype('i2')
    return ds_mask.chunk({'k': 1, 'face': -1})


def write_masks_to_zarr(ds_mask, output_path):
    import lzma
    lzma_filters = [dict(id=lzma.FILTER_DELTA, dist=1),
                    dict(id=lzma.FILTER_LZMA2, preset=1)]
    from numcodecs import LZMA
    compressor = LZMA(filters=lzma_filters, format=lzma.FORMAT_RAW)
    encoding = {vname: {'compressor': compressor} for vname in ds_mask.data_vars}
    return ds_mask.to_zarr(output_path, encoding=encoding)


def copy_masks_from_mds_to_zarr(grid_dir, output_path):
    ds_mask = load_masks_from_mds(grid_dir)
    write_masks_to_zarr(ds_mask, output_path)


_facet_strides = ((0,3), (3,6), (6,7), (7,10), (10,13))
def face_mask_to_facet_index_list(mask):
    nk, nf, nx, ny = mask.shape
    assert nf == 13
    index = np.asarray(mask.sum(axis=(2, 3)))
    index_facets = np.array([list(index[:, slice(*stride)].sum(axis=1))
                             for stride in _facet_strides]).transpose()
    return [0] + list(index_facets.ravel().cumsum())



from packaging import version

try:
    import fsspec
    assert version.parse(fsspec.__version__) >= version.parse("0.4.4")
except (ImportError, AssertionError): # pramga: no cover
    raise ImportError('The llcreader module requires fsspec version 0.4.4 or '
                      'or greater to be installed. '
                      'To install it, run `pip install fsspec`. See '
                      'https://filesystem-spec.readthedocs.io/en/latest/ '
                      'for more info.')

try:
    import zarr
    assert version.parse(zarr.__version__) >= version.parse("2.3.1")
except (ImportError, AssertionError): # pramga: no cover
    raise ImportError('The llcreader module requires zarr version 2.3.1 or '
                      'or greater to be installed. '
                      'To install it, run `pip install zarr`. See '
                      'https://zarr.readthedocs.io/en/stable/index.html#installation '
                      'for more info.')


from .known_models import *
from .stores import *
from .llcmodel import BaseLLCModel, faces_dataset_to_latlon



import re
import pytest
from dask.array.core import Array as dsa

llcreader = pytest.importorskip("xmitgcm.llcreader")

from .test_xmitgcm_common import llc_mds_datadirs

EXPECTED_VARS = ['Eta', 'KPPhbl', 'oceFWflx', 'oceQnet', 'oceQsw', 'oceSflux',
            'oceTAUX', 'oceTAUY', 'PhiBot', 'Salt', 'SIarea', 'SIheff',
            'SIhsalt', 'SIhsnow', 'SIuice', 'SIvice', 'Theta', 'U', 'V', 'W']
GRID_VARNAMES = ['AngleCS', 'AngleSN', 'DRC', 'DRF', 'DXC', 'DXG', 'DYC', 'DYG',
                 'Depth', 'PHrefC', 'PHrefF', 'RAC', 'RAS', 'RAW', 'RAZ', 'RC', 'RF',
                 'RhoRef', 'XC', 'XG', 'YC', 'YG', 'hFacC', 'hFacS', 'hFacW']


EXPECTED_COORDS = {2160: ['CS','SN','Depth',
                          'drC','drF','dxC','dxF','dxG','dyC','dyF','dyG',
                          'hFacC','hFacS','hFacW','PHrefC','PHrefF','rA','rAs','rAw',
                          'Z','Zp1','Zl','Zu','rhoRef','rLowC','rLowS','rLowW',
                          'rSurfC','rSurfS','rSurfW','XC','YC','rAz','XG','YG',
                          'dxV','dyU'],
                   4320: ['CS','SN','Depth',
                          'drC','drF','dxC','dxF','dxG','dyC','dyF','dyG',
                          'hFacC','hFacS','hFacW','PHrefC','PHrefF',
                          'rA','rAs','rAw','rhoRef','Z','Zp1','Zl','Zu','XC','YC',
                          'rAz','XG','YG','dxV','dyU'],
                   'aste_270': ["CS", "Depth", "PHrefC", "PHrefF", "SN",
                                "XC", "XG", "YC", "YG", "Z",
                                "Zl", "Zp1", "Zu", "drC", "drF",
                                "dxC", "dxG", "dyC", "dyG", "hFacC",
                                "hFacS", "hFacW", "maskC", "maskCtrlC", "maskCtrlS",
                                "maskCtrlW", "maskInC", "maskInS", "maskInW", "maskS",
                                "maskW", "niter", "rA", "rAs", "rAw",
                                "rAz", "rhoRef"]}

########### Generic llcreader tests on local data ##############################

@pytest.fixture(scope='module')
def local_llc90_store(llc_mds_datadirs):
    from fsspec.implementations.local import LocalFileSystem
    dirname, expected = llc_mds_datadirs
    fs = LocalFileSystem()
    store = llcreader.BaseStore(fs, base_path=dirname, grid_path=dirname)
    return store

@pytest.fixture(scope='module')
def llc90_kwargs():
    return dict(varnames=['S', 'T', 'U', 'V', 'Eta'],
                iter_start=0, iter_stop=9, iter_step=8)

def test_llc90_local_faces(local_llc90_store, llc90_kwargs):
    store = local_llc90_store
    model = llcreader.LLC90Model(store)
    ds_faces = model.get_dataset(**llc90_kwargs)
    assert set(llc90_kwargs['varnames']) == set(ds_faces.data_vars)
    assert ds_faces.dims == {'face': 13, 'i': 90, 'i_g': 90, 'j': 90, 'j_g': 90,
                             'k': 50, 'k_u': 50, 'k_l': 50, 'k_p1': 51, 'time': 2}

def test_llc90_dim_metadata(local_llc90_store, llc90_kwargs):
    store = local_llc90_store
    model = llcreader.LLC90Model(store)
    ds_faces = model.get_dataset(**llc90_kwargs)
    assert ds_faces.i.attrs['axis'] == 'X'

def test_llc90_local_latlon(local_llc90_store, llc90_kwargs):
    store = local_llc90_store
    model = llcreader.LLC90Model(store)
    ds_latlon = model.get_dataset(type='latlon', **llc90_kwargs)
    assert set(llc90_kwargs['varnames']) == set(ds_latlon.data_vars)
    assert ds_latlon.dims == {'i': 360, 'time': 2, 'k_p1': 51, 'face': 13,
                              'i_g': 360, 'k_u': 50, 'k': 50, 'k_l': 50,
                              'j_g': 270, 'j': 270}


# includes regression test for https://github.com/MITgcm/xmitgcm/issues/233
@pytest.mark.parametrize('rettype', ['faces', 'latlon'])
@pytest.mark.parametrize('k_levels, kp1_levels, k_chunksize',
        [(None, None, 1),
         ([1], [1, 2], 1),
         ([0, 2, 7, 9, 10, 20],
          [0,1,2,3,7,8,9,10,11,20,21], 1),
         ([0, 2, 7, 9, 10, 20],
          [0,1,2,3,7,8,9,10,11,20,21], 2)
         ])
@pytest.mark.parametrize('read_grid', [False, True]
)
def test_llc90_local_faces_load(local_llc90_store, llc90_kwargs, rettype, k_levels,
                                kp1_levels, k_chunksize, read_grid):
    store = local_llc90_store
    model = llcreader.LLC90Model(store)
    model.grid_varnames = GRID_VARNAMES
    ds = model.get_dataset(k_levels=k_levels, k_chunksize=k_chunksize,
                           type=rettype, read_grid=read_grid, **llc90_kwargs)
    if read_grid:
        # doesn't work because the variables change name
        # assert set(GRID_VARNAMES).issubset(set(ds.coords))
        pass
    if k_levels is None:
        assert list(ds.k.values) == list(range(50))
        assert list(ds.k_p1.values) == list(range(51))
    else:
        assert list(ds.k.values) == k_levels
        assert list(ds.k_p1.values) == kp1_levels
    assert all([cs==k_chunksize for cs in ds['T'].data.chunks[1]])

    ds.load()


@pytest.mark.parametrize('varname', [['U'], ['V']])
def test_vector_mate_error(local_llc90_store, varname):
    store = local_llc90_store
    model = llcreader.LLC90Model(store)
    with pytest.raises(ValueError, match=r".* must also be .*"):
        ds_latlon = model.get_dataset(type='latlon', varnames=varname, iter_start=0, iter_stop=9, iter_step=8)


########### ECCO Portal Tests ##################################################

@pytest.fixture(scope='module', params=[2160, 4320])
def ecco_portal_model(request):
    if request.param==2160:
        return llcreader.ECCOPortalLLC2160Model()
    else:
        return llcreader.ECCOPortalLLC4320Model()

def test_ecco_portal_faces(ecco_portal_model):
    # just get three timesteps
    iter_stop = ecco_portal_model.iter_start + 2 * ecco_portal_model.iter_step + 1
    ds_faces = ecco_portal_model.get_dataset(iter_stop=iter_stop)
    nx = ecco_portal_model.nx
    assert ds_faces.dims == {'face': 13, 'i': nx, 'i_g': nx, 'j': nx,
                              'j_g': nx, 'k': 90, 'k_u': 90, 'k_l': 90,
                              'k_p1': 91, 'time': 3}
    assert set(EXPECTED_VARS) == set(ds_faces.data_vars)
    assert set(EXPECTED_COORDS[nx]).issubset(set(ds_faces.coords))

    # make sure vertical coordinates are in one single chunk
    for fld in ds_faces[['Z','Zl','Zu','Zp1']].coords:
        if isinstance(ds_faces[fld].data,dsa):
            assert len(ds_faces[fld].data.chunks)==1
            assert (len(ds_faces[fld]),)==ds_faces[fld].data.chunks[0]


def test_ecco_portal_iterations(ecco_portal_model):
    with pytest.warns(RuntimeWarning, match=r"Iteration .* may not exist, you may need to change 'iter_start'"):
        ecco_portal_model.get_dataset(varnames=['Eta'], iter_start=ecco_portal_model.iter_start + 1, read_grid=False)

    with pytest.warns(RuntimeWarning, match=r"'iter_step' is not a multiple of .*, meaning some expected timesteps may not be returned"):
        ecco_portal_model.get_dataset(varnames=['Eta'], iter_step=ecco_portal_model.iter_step - 1, read_grid=False)

    with pytest.warns(RuntimeWarning, match=r"Some requested iterations may not exist, you may need to change 'iters'"):
        iters = [ecco_portal_model.iter_start, ecco_portal_model.iter_start + 1]
        ecco_portal_model.get_dataset(varnames=['Eta'], iters=iters, read_grid=False)

    with pytest.warns(None) as record:
        ecco_portal_model.get_dataset(varnames=['Eta'], read_grid=False)
    assert not record


@pytest.mark.slow
def test_ecco_portal_load(ecco_portal_model):
    # an expensive test because it actually loads data
    iter_stop = ecco_portal_model.iter_start + 2 * ecco_portal_model.iter_step + 1
    ds_faces = ecco_portal_model.get_dataset(varnames=['Eta'], iter_stop=iter_stop)
    # a lookup table
    expected = {2160: -1.3054643869400024, 4320: -1.262018084526062}
    assert ds_faces.Eta[0, 0, -1, -1].values.item() == expected[ecco_portal_model.nx]

def test_ecco_portal_latlon(ecco_portal_model):
    iter_stop = ecco_portal_model.iter_start + 2 * ecco_portal_model.iter_step + 1
    ds_ll = ecco_portal_model.get_dataset(iter_stop=iter_stop, type='latlon')
    nx = ecco_portal_model.nx
    assert ds_ll.dims == {'i': 4*nx, 'k_u': 90, 'k_l': 90, 'time': 3,
                             'k': 90, 'j_g': 3*nx, 'i_g': 4*nx, 'k_p1': 91,
                             'j': 3*nx, 'face': 13}
    assert set(EXPECTED_VARS) == set(ds_ll.data_vars)
    assert set(EXPECTED_COORDS[nx]).issubset(set(ds_ll.coords))

    # make sure vertical coordinates are in one single chunk
    for fld in ds_ll[['Z','Zl','Zu','Zp1']].coords:
        if isinstance(ds_ll[fld].data,dsa):
            assert len(ds_ll[fld].data.chunks)==1
            assert (len(ds_ll[fld]),)==ds_ll[fld].data.chunks[0]


########### ASTE Portal Tests ##################################################
@pytest.fixture(scope='module')
def aste_portal_model():
    return llcreader.CRIOSPortalASTE270Model()

def test_aste_portal_faces(aste_portal_model):
    # just get three timesteps
    iters = aste_portal_model.iters[:3]
    ds_faces = aste_portal_model.get_dataset(iters=iters)
    nx = aste_portal_model.nx
    assert ds_faces.dims == {'face': 6, 'i': nx, 'i_g': nx, 'j': nx,
                              'j_g': nx, 'k': 50, 'k_u': 50, 'k_l': 50,
                              'k_p1': 51, 'time': 3}
    assert set(aste_portal_model.varnames) == set(ds_faces.data_vars)
    assert set(EXPECTED_COORDS['aste_270']).issubset(set(ds_faces.coords))

    # make sure vertical coordinates are in one single chunk
    for fld in ds_faces[['Z','Zl','Zu','Zp1']].coords:
        if isinstance(ds_faces[fld].data,dsa):
            assert len(ds_faces[fld].data.chunks)==1
            assert (len(ds_faces[fld]),)==ds_faces[fld].data.chunks[0]


def test_aste_portal_iterations(aste_portal_model):
    with pytest.warns(RuntimeWarning, match=r"Some requested iterations may not exist, you may need to change 'iters'"):
        #iters = [ecco_portal_model.iter_start, ecco_portal_model.iter_start + 1]
        iters = aste_portal_model.iters[:2]
        iters[1] = iters[1] + 1
        aste_portal_model.get_dataset(varnames=['ETAN'], iters=iters, read_grid=False)

    with pytest.warns(None) as record:
        iters = aste_portal_model.iters[:2]
        aste_portal_model.get_dataset(varnames=['ETAN'], iters=iters, read_grid=False)
    assert not record


@pytest.mark.slow
def test_aste_portal_load(aste_portal_model):
    # an expensive test because it actually loads data
    iters = aste_portal_model.iters[:3]
    ds_faces = aste_portal_model.get_dataset(varnames=['ETAN'], iters=iters)
    expected = 0.641869068145752
    assert ds_faces.ETAN[0, 1, 0, 0].values.item() == expected

def test_aste_portal_latlon(aste_portal_model):
    iters = aste_portal_model.iters[:3]
    with pytest.raises(TypeError):
        ds_ll = aste_portal_model.get_dataset(iters=iters,type='latlon')





import pytest
import os
import fnmatch
import tarfile
import numpy as np
import dask
from contextlib import contextmanager
import py
import tempfile
import hashlib
try:
    import urllib.request as req
except ImportError:
    # urllib in python2 has different structure
    import urllib as req

from xmitgcm.file_utils import clear_cache

@contextmanager
def hide_file(origdir, *basenames):
    """Temporarily hide files within the context."""
    # make everything a py.path.local
    tmpdir = py.path.local(tempfile.mkdtemp())
    origdir = py.path.local(origdir)
    oldpaths = [origdir.join(basename) for basename in basenames]
    newpaths = [tmpdir.join(basename) for basename in basenames]

    # move the files
    for oldpath, newpath in zip(oldpaths, newpaths):
        oldpath.rename(newpath)
    # clear the cache if it exists
    clear_cache()
    try:
        yield str(tmpdir)
    finally:
        # move them back
        for oldpath, newpath in zip(oldpaths, newpaths):
            newpath.rename(oldpath)


dlroot = 'https://ndownloader.figshare.com/files/'

# parameterized fixture are complicated
# http://docs.pytest.org/en/latest/fixture.html#fixture-parametrize

# dictionary of archived experiments and some expected properties
_experiments = {
    'global_oce_latlon': {'geometry': 'sphericalpolar',
                          'dlink': dlroot + '14066630',
                          'md5': '0a846023d01cbcc16bed4963431968cf',
                          'shape': (15, 40, 90), 'test_iternum': 39600,
                          'expected_values': {'XC': ((0, 0), 2)},
                          'dtype': np.dtype('f4'),
                          'layers': {'1RHO': 31},
                          'diagnostics': ('DiagGAD-T',
                                          ['TOTTTEND', 'ADVr_TH',
                                           'ADVx_TH', 'ADVy_TH',
                                           'DFrE_TH', 'DFxE_TH',
                                           'DFyE_TH', 'DFrI_TH',
                                           'UTHMASS', 'VTHMASS', 'WTHMASS'])},
    'barotropic_gyre': {'geometry': 'cartesian',
                        'dlink': dlroot + '14066618',
                        'md5': '5200149791bfd24989ad8b98c18937dc',
                        'shape': (1, 60, 60), 'test_iternum': 10,
                        'dtype': np.dtype('f4'),
                        'expected_values': {'XC': ((0, 0), 10000.0)},
                        'all_iters': [0, 10],
                        'delta_t': 3600,
                        'prefixes': ['T', 'S', 'Eta', 'U', 'V', 'W']},
    'internal_wave': {'geometry': 'sphericalpolar',
                      'dlink': dlroot + '14066642',
                      'md5': 'eedfab1aec365fd8c17d3bc0f86a1431',
                      'shape': (20, 1, 30), 'test_iternum': 100,
                      'dtype': np.dtype('f8'),
                      'expected_values': {'XC': ((0, 0), 109.01639344262296)},
                      'all_iters': [0, 100, 200],
                      'ref_date': "1990-1-1",
                      'delta_t': 60,
                      'expected_time': [
                          (0, np.datetime64('1990-01-01T00:00:00.000000000')),
                          (1, np.datetime64('1990-01-01T01:40:00.000000000'))],
                      # these diagnostics won't load because not all levels
                      # were output...no idea how to overcome that bug
                      # 'diagnostics': ('diagout1', ['UVEL', 'VVEL']),
                      'prefixes': ['T', 'S', 'Eta', 'U', 'V', 'W']},
    'global_oce_llc90': {'geometry': 'llc',
                         'dlink': dlroot + '14066567',
                         'md5': '6c309416f91ae9baaf1fb21b3dc50e49',
                         'ref_date': "1948-01-01 12:00:00",
                         'delta_t': 3600,
                         'expected_time': [
                             (0, np.datetime64('1948-01-01T12:00:00.000000000')),
                             (1, np.datetime64('1948-01-01T20:00:00.000000000'))],
                         'shape': (50, 13, 90, 90), 'test_iternum': 8,
                         'dtype': np.dtype('f4'),
                         'expected_values': {'XC': ((2, 3, 5), -32.5)},
                         'diagnostics': ('state_2d_set1', ['ETAN',
                                                           'SIarea',
                                                           'SIheff',
                                                           'SIhsnow',
                                                           'DETADT2',
                                                           'PHIBOT',
                                                           'sIceLoad',
                                                           'MXLDEPTH',
                                                           'oceSPDep',
                                                           'SIatmQnt',
                                                           'SIatmFW',
                                                           'oceQnet',
                                                           'oceFWflx',
                                                           'oceTAUX',
                                                           'oceTAUY',
                                                           'ADVxHEFF',
                                                           'ADVyHEFF',
                                                           'DFxEHEFF',
                                                           'DFyEHEFF',
                                                           'ADVxSNOW',
                                                           'ADVySNOW',
                                                           'DFxESNOW',
                                                           'DFyESNOW',
                                                           'SIuice',
                                                           'SIvice'])},
    'curvilinear_leman': {'geometry': 'curvilinear',
                          'dlink': dlroot + '14066621',
                          'md5': 'c3203ae1fb0d6a61174dd680b7660894',
                          'delta_t': 20,
                          'ref_date': "2013-11-12 12:00",
                          'shape': (35, 64, 340),
                          'test_iternum': 6,
                          'dtype': np.dtype('f4'),
                          'expected_values': {'XC': ((0, 0), 501919.21875)},
                          'all_iters': [0, 3, 6],
                          'expected_time': [
                              (0, np.datetime64('2013-11-12T12:00:00.000000000')),
                              (1, np.datetime64('2013-11-12T12:02:00.000000000'))],
                          'prefixes': ['THETA']},
    'ideal_2D_oce': {'geometry': 'sphericalpolar',
                     'dlink': dlroot + '17288255',
                     'md5': 'd8868731ff6a8fd951babefbc5ea69ba',
                     'expected_namelistvals': {'eosType': 'LINEAR',
                                               'viscAh': 12e5,
                                               'niter0': 36000,
                                               'delX': [3.],
                                               'fileName': ['surfDiag', 'dynDiag',
                                                            'oceDiag', 'flxDiag'],
                                               'levels': [[1.0], [],
                                                          [2., 3., 4., 5., 6.,
                                                           7., 8., 9., 10., 11.,
                                                           12., 13.]],
                                               'useDiagnostics': True},
                     'diag_levels': {'surfDiag': ([1], (0, -50)),
                                     'oceDiag': (slice(2, 14), (0, -122.5))},
                     'expected_values': {'XC': ((0, 0), 1.5)},
                     'shape': (15, 56, 1),
                     'test_iternum': 36020,
                     'dtype': np.dtype('f4')}
}


_grids = {
    'grid_llc90': {'geometry': 'llc', 'domain': 'llc',
                   'dlink': dlroot + '14072594',
                   'md5': 'f66c3195a62790d539debe6ca8f3a851',
                   'gridfile': 'tile<NFACET>.mitgrid',
                   'nx': 90, 'shape': (13, 90, 90)},

    'grid_aste270': {'geometry': 'llc', 'domain': 'aste',
                     'dlink': dlroot + '14072591',
                     'md5': '92b28c65e0dfb54b253bfcd0a249359b',
                     'gridfile': 'tile<NFACET>.mitgrid',
                     'nx': 270, 'shape': (6, 270, 270)}  # ,

    #    'grid_cs32': {'geometry': 'cs', 'domain': 'cs',
    #                  'dlink': dlroot + '14072597',
    #                  'md5': '848cd5b6daab5b069e96a0cff67d4b57',
    #                  'gridfile': 'grid_cs32.face<NFACET>.bin',
    #                  'nx': 32, 'shape': (6, 32, 32)}
}


def setup_mds_dir(tmpdir_factory, request, db):
    """Helper function for setting up test cases."""
    expt_name = request.param
    expected_results = db[expt_name]
    target_dir = str(tmpdir_factory.mktemp('mdsdata'))
    try:
        # user-defined directory for test datasets
        data_dir = os.environ["XMITGCM_TESTDATA"]
    except KeyError:
        # default to HOME/.xmitgcm-test-data/
        data_dir = os.environ["HOME"] + '/.xmitgcm-test-data'
    # create the directory if it doesn't exixt
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    datafile = os.path.join(data_dir, expt_name + '.tar.gz')
    # download if does not exist locally
    if not os.path.exists(datafile):
        print('File does not exist locally, downloading...')
        download_archive(expected_results['dlink'], datafile)
        localmd5 = file_md5_checksum(datafile)
        if localmd5 != expected_results['md5']:
            os.remove(datafile)
            msg = """
            MD5 checksum does not match, try downloading dataset again.
            """
            raise IOError(msg)

    return untar(data_dir, expt_name, target_dir), expected_results


def download_archive(url, filename):
    """ download file from url into datafile

    PARAMETERS:

    url: str
        url to retrieve
    filename: str
        file to save on disk
    """

    req.urlretrieve(url, filename)
    return None


def untar(data_dir, basename, target_dir):
    """Unzip a tar file into the target directory. Return path to unzipped
    directory."""
    datafile = os.path.join(data_dir, basename + '.tar.gz')
    if not os.path.exists(datafile):
        raise IOError('Could not find data file %s' % datafile)
    tar = tarfile.open(datafile)
    tar.extractall(target_dir)
    tar.close()
    # subdirectory where file should have been untarred.
    # assumes the directory is the same name as the tar file itself.
    # e.g. testdata.tar.gz --> testdata/
    fulldir = os.path.join(target_dir, basename)
    if not os.path.exists(fulldir):
        raise IOError('Could not find tar file output dir %s' % fulldir)
    # the actual data lives in a file called testdata
    # clean up ugly weird hidden files that mac-os sometimes puts in the archive
    # https://unix.stackexchange.com/questions/9665/create-tar-archive-of-a-directory-except-for-hidden-files
    # https://superuser.com/questions/259703/get-mac-tar-to-stop-putting-filenames-in-tar-archives
    bad_files = [f for f in os.listdir(fulldir)
                 if fnmatch.fnmatch(f, '._*') ]
    for f in bad_files:
        os.remove(os.path.join(fulldir, f))

    return fulldir


def file_md5_checksum(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        hash_md5.update(f.read())
    return hash_md5.hexdigest()


# find the tar archive in the test directory
# http://stackoverflow.com/questions/29627341/pytest-where-to-store-expected-data
@pytest.fixture(scope='module', params=_experiments.keys())
def all_mds_datadirs(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)


@pytest.fixture(scope='module', params=['barotropic_gyre', 'internal_wave'])
def multidim_mds_datadirs(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)


@pytest.fixture(scope='module', params=['global_oce_latlon',
                                        'global_oce_llc90'])
def mds_datadirs_with_diagnostics(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)


@pytest.fixture(scope='module', params=['internal_wave', 'global_oce_llc90'])
def mds_datadirs_with_refdate(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)


@pytest.fixture(scope='module', params=['global_oce_latlon'])
def layers_mds_datadirs(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)


@pytest.fixture(scope='module', params=['global_oce_llc90'])
def llc_mds_datadirs(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)


@pytest.fixture(scope='module', params=_grids.keys())
def all_grid_datadirs(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _grids)


@pytest.fixture(scope='module', params=['ideal_2D_oce'])
def mds_datadirs_with_inputfiles(tmpdir_factory, request):
    return setup_mds_dir(tmpdir_factory, request, _experiments)



from xmitgcm.test.test_xmitgcm_common import *
import xmitgcm
import xarray as xr
from glob import glob
from shutil import copyfile
import py

_TESTDATA_FILENAME = 'testdata.tar.gz'
_TESTDATA_ITERS = [39600, ]
_TESTDATA_DELTAT = 86400

_EXPECTED_GRID_VARS = ['XC', 'YC', 'XG', 'YG', 'Zl', 'Zu', 'Z', 'Zp1', 'dxC',
                       'rAs', 'rAw', 'Depth', 'rA', 'dxG', 'dyG', 'rAz', 'dyC',
                       'PHrefC', 'drC', 'PHrefF', 'drF',
                       'hFacS', 'hFacC', 'hFacW']


# a meta test
def test_file_hiding(all_mds_datadirs):
    dirname, _ = all_mds_datadirs
    basenames = ['XC.data', 'XC.meta']
    for basename in basenames:
        assert os.path.exists(os.path.join(dirname, basename))
    with hide_file(dirname, *basenames):
        for basename in basenames:
            assert not os.path.exists(os.path.join(dirname, basename))
    for basename in basenames:
        assert os.path.exists(os.path.join(dirname, basename))

#########################################################
### Below are all tests that actually create datasets ###
#########################################################


def test_open_mdsdataset_minimal(all_mds_datadirs):
    """Create a minimal xarray object with only dimensions in it."""

    dirname, expected = all_mds_datadirs

    ds = xmitgcm.open_mdsdataset(
        dirname, iters=None, read_grid=False, swap_dims=False,
        geometry=expected['geometry'])

    # the expected dimensions of the dataset
    eshape = expected['shape']
    if len(eshape) == 3:
        nz, ny, nx = eshape
        nface = None
    elif len(eshape) == 4:
        nz, nface, ny, nx = eshape
    else:
        raise ValueError("Invalid expected shape")
    coords = {'i': np.arange(nx),
              'i_g': np.arange(nx),
              # 'i_z': np.arange(nx),
              'j': np.arange(ny),
              'j_g': np.arange(ny),
              # 'j_z': np.arange(ny),
              'k': np.arange(nz),
              'k_u': np.arange(nz),
              'k_l': np.arange(nz),
              'k_p1': np.arange(nz+1)}
    if nface is not None:
        coords['face'] = np.arange(nface)

    if 'layers' in expected:
        for layer_name, n_layers in expected['layers'].items():
            for suffix, offset in zip(['bounds', 'center', 'interface'],
                                      [0, -1, -2]):
                dimname = 'l' + layer_name[0] + '_' + suffix[0]
                index = np.arange(n_layers + offset)
                coords[dimname] = index

    ds_expected = xr.Dataset(coords=coords)
    assert ds_expected.equals(ds)

    # check that the datatypes are correct
    for key in coords:
        assert ds[key].dtype == np.int64

    # check for comodo metadata needed by xgcm
    assert ds['i'].attrs['axis'] == 'X'
    assert ds['i_g'].attrs['axis'] == 'X'
    assert ds['i_g'].attrs['c_grid_axis_shift'] == -0.5
    assert ds['j'].attrs['axis'] == 'Y'
    assert ds['j_g'].attrs['axis'] == 'Y'
    assert ds['j_g'].attrs['c_grid_axis_shift'] == -0.5
    assert ds['k'].attrs['axis'] == 'Z'
    assert ds['k_l'].attrs['axis'] == 'Z'
    assert ds['k_l'].attrs['c_grid_axis_shift'] == -0.5


def test_read_grid(all_mds_datadirs):
    """Make sure we read all the grid variables."""
    dirname, expected = all_mds_datadirs
    ds = xmitgcm.open_mdsdataset(
        dirname, iters=None, read_grid=True,
        geometry=expected['geometry'])

    for vname in _EXPECTED_GRID_VARS:
        assert vname in ds.variables

    # make sure angle is present
    if expected['geometry'] in ['llc', 'curvilinear']:
        assert 'CS' in ds.coords
        assert 'SN' in ds.coords

    # actually load the data, to check for dask-related errors
    ds.load()

def test_values_and_endianness(all_mds_datadirs):
    """Make sure we read all the grid variables."""
    dirname, expected = all_mds_datadirs

    if expected['geometry']=='llc' and (dask.__version__ < '0.11.2'):
        pytest.xfail("LLC value tests require fixed dask")

    # default endianness
    ds = xmitgcm.open_mdsdataset(
                dirname, iters=None, read_grid=True, swap_dims=False,
                geometry=expected['geometry'])
    # now reverse endianness
    ds_le = xmitgcm.open_mdsdataset(
                dirname, iters=None, read_grid=True, endian='<',
                swap_dims=False,
                geometry=expected['geometry'])

    for vname, (idx, val) in expected['expected_values'].items():
        np.testing.assert_allclose(ds[vname].values[idx], val)
        # dask arrays that have been concatenated revert to native endianness
        # https://github.com/dask/dask/issues/1647
        if ds[vname].dtype.byteorder=='>':
            val_le = ds[vname].values.newbyteorder('<')[idx]
            np.testing.assert_allclose(ds_le[vname].values[idx], val_le)

def test_open_dataset_no_meta(all_mds_datadirs):
    """Make sure we read  variables with no .meta files."""
    dirname, expected = all_mds_datadirs

    shape = expected['shape']

    nz = shape[0]
    ny, nx = shape[-2:]
    shape_2d = shape[1:]
    dims_2d = ('j', 'i')
    if expected['geometry']=='llc':
        dims_2d = ('face',) + dims_2d
        ny = nx*shape[-3]
    dims_3d = dims_2d if nz==1 else ('k',) + dims_2d
    dims_2d = ('time',) + dims_2d
    dims_3d = ('time',) + dims_3d

    it = expected['test_iternum']
    kwargs = dict(iters=it, geometry=expected['geometry'], read_grid=False,
                  swap_dims=False, default_dtype=expected['dtype'])

    # a 3D file
    to_hide = ['T.%010d.meta' % it, 'Eta.%010d.meta' % it]
    with hide_file(dirname, *to_hide):
        ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'], **kwargs)
        assert ds['T'].dims == dims_3d
        assert ds['T'].values.ndim == len(dims_3d)
        assert ds['Eta'].dims == dims_2d
        assert ds['Eta'].values.ndim == len(dims_2d)

        with pytest.raises(IOError):
            xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'], iters=it,
                                    geometry=expected['geometry'],
                                    read_grid=False)
            pytest.fail("Expecting IOError when default_dtype "
                        "is not precised (i.e., None)")

    # now get rid of the variables used to infer dimensions
    with hide_file(dirname, 'XC.meta', 'RC.meta'):
        with pytest.raises(IOError):
            ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'], **kwargs)
        if expected['geometry']=='llc':
            ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'],
                                         nx=nx, nz=nz, **kwargs)
            with hide_file(dirname, *to_hide):
                ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'],
                                             nx=nx, nz=nz, **kwargs)
        else:
            ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'],
                                         nx=nx, ny=ny, nz=nz, **kwargs)
            with hide_file(dirname, *to_hide):
                ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'],
                                             nx=nx, ny=ny, nz=nz, **kwargs)

    # try just hiding RC
    with hide_file(dirname, 'RC.meta'):
        if expected['geometry']=='llc':
            ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'],
                                         nz=nz, **kwargs)
        else:
            ds = xmitgcm.open_mdsdataset(dirname, prefix=['T', 'Eta'],
                                         nz=nz, **kwargs)

def test_open_dataset_2D_diags(all_mds_datadirs):
    # convert 3D fields with only 2D diagnostic output
    # https://github.com/xgcm/xmitgcm/issues/140
    dirname, expected = all_mds_datadirs

    shape = expected['shape']

    nz = shape[0]
    ny, nx = shape[-2:]
    shape_2d = shape[1:]
    dims_2d = ('j', 'i')
    if expected['geometry']=='llc':
        dims_2d = ('face',) + dims_2d
        ny = nx*shape[-3]
    dims_3d = dims_2d if nz==1 else ('k',) + dims_2d
    dims_2d = ('time',) + dims_2d
    dims_3d = ('time',) + dims_3d

    it = expected['test_iternum']
    kwargs = dict(iters=it, geometry=expected['geometry'], read_grid=False,
                  swap_dims=False)

    to_hide = ['T.%010d.meta' % it, 'T.%010d.data' % it]
    with hide_file(dirname, *to_hide):

        ldir = py.path.local(dirname)
        old_prefix = 'Eta.%010d' % it
        new_prefix = 'T.%010d' % it
        for suffix in ['.data', '.meta']:
            lp = ldir.join(old_prefix + suffix)
            lp.copy(ldir.join(new_prefix + suffix))

        ds = xmitgcm.open_mdsdataset(dirname, prefix=['T'], **kwargs)

def test_swap_dims(all_mds_datadirs):
    """See if we can swap dimensions."""

    dirname, expected = all_mds_datadirs
    kwargs = dict(iters=None, read_grid=True, geometry=expected['geometry'])

    expected_dims = ['XC', 'XG', 'YC', 'YG', 'Z', 'Zl', 'Zp1', 'Zu']

    # make sure we never swap if not reading grid
    assert 'i' in xmitgcm.open_mdsdataset(dirname,
        iters=None, read_grid=False, geometry=expected['geometry'])
    if expected['geometry'] in ('llc', 'curvilinear'):
        # make sure swapping is not the default
        ds = xmitgcm.open_mdsdataset(dirname, **kwargs)
        assert 'i' in ds
        # and is impossible
        with pytest.raises(ValueError) as excinfo:
            ds = xmitgcm.open_mdsdataset(
                        dirname, swap_dims=True, **kwargs)
    else:
        # make sure swapping *IS* the default
        assert 'i' not in xmitgcm.open_mdsdataset(dirname, **kwargs)
        ds = xmitgcm.open_mdsdataset(
                    dirname, geometry=expected['geometry'],
                    iters=None, read_grid=True, swap_dims=True,
                    grid_vars_to_coords=True)


        # check for comodo metadata needed by xgcm
        assert ds['XC'].attrs['axis'] == 'X'
        assert ds['XG'].attrs['axis'] == 'X'
        assert ds['XG'].attrs['c_grid_axis_shift'] == -0.5
        assert ds['YC'].attrs['axis'] == 'Y'
        assert ds['YG'].attrs['axis'] == 'Y'
        assert ds['YG'].attrs['c_grid_axis_shift'] == -0.5
        assert ds['Z'].attrs['axis'] == 'Z'
        assert ds['Zl'].attrs['axis'] == 'Z'
        assert ds['Zl'].attrs['c_grid_axis_shift'] == -0.5

        # add extra layers dimensions if needed
        if 'layers' in expected:
            for layer_name in expected['layers']:
                extra_dims = ['layer_' + layer_name + suffix for suffix in
                              ['_bounds', '_center', '_interface']]
                expected_dims += extra_dims

        assert list(ds.dims.keys()) == expected_dims

        # make sure swapping works with multiple iters
        ds = xmitgcm.open_mdsdataset(dirname, geometry=expected['geometry'],
                                     prefix=['S'])
        #print(ds)
        ds.load()
        assert 'XC' in ds['S'].dims
        assert 'YC' in ds['S'].dims



def test_prefixes(all_mds_datadirs):
    """Make sure we read all the grid variables."""

    dirname, expected = all_mds_datadirs
    prefixes = ['U', 'V', 'W', 'T', 'S', 'PH']  # , 'PHL', 'Eta']
    iters = [expected['test_iternum']]
    ds = xmitgcm.open_mdsdataset(
                dirname, iters=iters, prefix=prefixes,
                read_grid=False, geometry=expected['geometry'])

    for p in prefixes:
        assert p in ds

def test_separate_grid_dir(all_mds_datadirs):
    """Make sure we can have the grid files in a separate directory."""

    dirname, expected = all_mds_datadirs
    prefixes = ['U', 'V', 'W', 'T', 'S', 'PH']  # , 'PHL', 'Eta']
    iters = [expected['test_iternum']]

    with hide_file(dirname,
                    *['XC.meta', 'XC.data', 'RC.meta', 'RC.data']) as grid_dir:
        ds = xmitgcm.open_mdsdataset(
                dirname, grid_dir=grid_dir, iters=iters, prefix=prefixes,
                read_grid=False, geometry=expected['geometry'])
        for p in prefixes:
            assert p in ds

def test_multiple_iters(multidim_mds_datadirs):
    """Test ability to load multiple iters into a single dataset."""

    dirname, expected = multidim_mds_datadirs
    # first try specifying the iters
    ds = xmitgcm.open_mdsdataset(
        dirname, read_grid=False, geometry=expected['geometry'],
        iters=expected['all_iters'],
        prefix=expected['prefixes'])
    assert list(ds.iter.values) == expected['all_iters']

    # now infer the iters, should be the same
    ds2 = xmitgcm.open_mdsdataset(
        dirname, read_grid=False, iters='all', geometry=expected['geometry'],
        prefix=expected['prefixes'])
    assert ds.equals(ds2)

    # In the test datasets, there is no PHL.0000000000.data file.
    # By default we infer the prefixes from the first iteration number, so this
    # leads to an error.
    # (Need to specify iters because there is some diagnostics output with
    # weird iterations numbers present in some experiments.)
    with pytest.raises(IOError):
        ds = xmitgcm.open_mdsdataset(
            dirname, read_grid=False, iters=expected['all_iters'],
            geometry=expected['geometry'])

    # now hide all the PH and PHL files: should be able to infer prefixes fine
    missing_files = [os.path.basename(f)
                     for f in glob(os.path.join(dirname, 'PH*.0*data'))]
    print(missing_files)
    with hide_file(dirname, *missing_files):
        ds = xmitgcm.open_mdsdataset(
            dirname, read_grid=False, iters=expected['all_iters'],
            geometry=expected['geometry'])


def test_date_parsing(mds_datadirs_with_refdate):
    """Verify that time information is decoded properly."""
    dirname, expected = mds_datadirs_with_refdate

    ds = xmitgcm.open_mdsdataset(dirname, iters='all', prefix=['S'],
                              ref_date=expected['ref_date'], read_grid=False,
                              delta_t=expected['delta_t'],
                              geometry=expected['geometry'])

    for i, date in expected['expected_time']:
        assert ds.time[i].values == date

    # since time was decoded, this encoding should be removed from attributes
    assert 'units' not in ds.time.attrs
    assert 'calendar' not in ds.time.attrs

def test_serialize_nonstandard_calendar(multidim_mds_datadirs, tmp_path):
    dirname, expected = multidim_mds_datadirs
    ref_date = '2680-01-01 00:00:00'
    calendar = '360_day'
    ds = xmitgcm.open_mdsdataset(dirname, iters='all', prefix=['S'],
                                 ref_date=ref_date,
                                 calendar=calendar,
                                 read_grid=False,
                                 delta_t=expected['delta_t'],
                                 geometry=expected['geometry'])
    ds.to_zarr(tmp_path / 'test.zarr')


def test_diagnostics(mds_datadirs_with_diagnostics):
    """Try reading dataset with diagnostics output."""
    dirname, expected = mds_datadirs_with_diagnostics

    diag_prefix, expected_diags = expected['diagnostics']
    ds = xmitgcm.open_mdsdataset(dirname,
                                              read_grid=False,
                                              iters=expected['test_iternum'],
                                              prefix=[diag_prefix],
                                              geometry=expected['geometry'])
    for diagname in expected_diags:
        assert diagname in ds
        # check vector mates
        if 'mate' in ds[diagname].attrs:
            mate = ds[diagname].attrs['mate']
            assert ds[mate].attrs['mate'] == diagname

def test_default_diagnostics(mds_datadirs_with_diagnostics):
    """Try reading dataset with diagnostics output."""
    dirname, expected = mds_datadirs_with_diagnostics

    diag_prefix, expected_diags = expected['diagnostics']
    with hide_file(dirname, 'available_diagnostics.log'):
        ds = xmitgcm.open_mdsdataset(dirname,
                                              read_grid=False,
                                              iters=expected['test_iternum'],
                                              prefix=[diag_prefix],
                                              geometry=expected['geometry'])
    for diagname in expected_diags:
        assert diagname in ds
        # check vector mates
        if 'mate' in ds[diagname].attrs:
            mate = ds[diagname].attrs['mate']
            assert ds[mate].attrs['mate'] == diagname

def test_avail_diags_in_grid_dir(mds_datadirs_with_diagnostics):
    """Try reading dataset with diagnostics output."""
    dirname, expected = mds_datadirs_with_diagnostics

    diag_prefix, expected_diags = expected['diagnostics']
    iters = expected['test_iternum']

    with hide_file(dirname,
                    *['XC.meta', 'XC.data', 'RC.meta', 'RC.data',
                      'available_diagnostics.log']) as grid_dir:
        ds = xmitgcm.open_mdsdataset(
                dirname, grid_dir=grid_dir, iters=iters, prefix=[diag_prefix],
                read_grid=False, geometry=expected['geometry'])

    for diagname in expected_diags:
        assert diagname in ds
        if 'mate' in ds[diagname].attrs:
            mate = ds[diagname].attrs['mate']
            assert ds[mate].attrs['mate'] == diagname

def test_layers_diagnostics(layers_mds_datadirs):
    """Try reading dataset with layers output."""
    dirname, expected = layers_mds_datadirs
    ds = xmitgcm.open_mdsdataset(dirname, iters='all', swap_dims=False,
                            geometry=expected['geometry'])
    layer_name = list(expected['layers'].keys())[0]
    layer_id = 'l' + layer_name[0]
    for suf in ['bounds', 'center', 'interface']:
        assert ('layer_' + layer_name + '_' + suf) in ds
        assert (layer_id + '_' + suf[0]) in ds.dims

    # a few random expected variables
    expected_vars = {'LaUH' + layer_name:
                     ('time', layer_id + '_c', 'j', 'i_g'),
                     'LaVH' + layer_name:
                     ('time', layer_id + '_c', 'j_g', 'i'),
                     'LaTs' + layer_name:
                     ('time', layer_id + '_i', 'j', 'i')}
    for var, dims in expected_vars.items():
        assert var in ds
        assert ds[var].dims == dims

@pytest.mark.parametrize("method", ["smallchunks"])
@pytest.mark.parametrize("with_refdate", [True, False])
def test_llc_dims(llc_mds_datadirs, method, with_refdate):
    """Check that the LLC file dimensions are correct."""
    dirname, expected = llc_mds_datadirs
    if with_refdate:
        kwargs = {'delta_t': expected['delta_t'],
                  'ref_date': expected['ref_date']}
    else:
        kwargs = {}
    ds = xmitgcm.open_mdsdataset(dirname,
                            iters=expected['test_iternum'],
                            geometry=expected['geometry'], llc_method=method,
                            **kwargs)

    nz, nface, ny, nx = expected['shape']
    nt = 1

    assert ds.dims['face'] == 13
    assert ds.rA.dims == ('face', 'j', 'i')
    assert ds.rA.values.shape == (nface, ny, nx)
    assert ds.U.dims == ('time', 'k', 'face', 'j', 'i_g')
    assert ds.U.values.shape == (nt, nz, nface, ny, nx)
    assert ds.V.dims == ('time', 'k', 'face', 'j_g', 'i')
    assert ds.V.values.shape == (nt, nz, nface, ny, nx)

    print(ds.U.chunks)
    if method == "smallchunks":
        assert ds.U.chunks == (nt*(1,), nz*(1,), nface*(1,), (ny,), (nx,))


def test_drc_length(all_mds_datadirs):
    """Test that open_mdsdataset is adding an extra level to drC if it has length nr"""
    dirname, expected = all_mds_datadirs
    # Only older versions of the gcm have len(drC) = nr, so force len(drC) = nr for the test
    copyfile(os.path.join(dirname, 'DRF.data'),
             os.path.join(dirname, 'DRC.data'))
    copyfile(os.path.join(dirname, 'DRF.meta'),
             os.path.join(dirname, 'DRC.meta'))
    ds = xmitgcm.open_mdsdataset(
        dirname, iters=None, read_grid=True,
        geometry=expected['geometry'])
    assert len(ds.drC) == (len(ds.drF)+1)


def test_mask_values(all_mds_datadirs):
    """Test that open_mdsdataset generates binary masks with correct values"""

    dirname, expected = all_mds_datadirs
    ds = xmitgcm.open_mdsdataset(
        dirname, iters=None, read_grid=True,
        geometry=expected['geometry'])

    hFac_list = ['hFacC', 'hFacW', 'hFacS']
    mask_list = ['maskC', 'maskW', 'maskS']

    for hFac, mask in zip(hFac_list, mask_list):
        xr.testing.assert_equal(ds[hFac] * ds[mask], ds[hFac])

#
# Series of tests which try to open a dataset with different combinations of
# of options, to identify if ref_date can trigger an error
#


@pytest.mark.parametrize("load", [True, False])
# can't call swap_dims without read_grid=True
@pytest.mark.parametrize("swap_dims, read_grid", [(True, True),
                                                  (False, True),
                                                  (False, False)])
def test_ref_date(mds_datadirs_with_refdate, swap_dims, read_grid, load):
    """With ref_date, without grid."""
    dirname, expected = mds_datadirs_with_refdate

    if expected['geometry']=='llc' and swap_dims:
        pytest.skip("can't swap_dims with geometry=='llc'")

    ds = xmitgcm.open_mdsdataset(dirname, iters='all', prefix=['S'],
                                 ref_date=expected['ref_date'],
                                 delta_t=expected['delta_t'],
                                 geometry=expected['geometry'],
                                 read_grid=read_grid, swap_dims=swap_dims)
    if load:
        ds.time.load()


@pytest.mark.parametrize("method", ["smallchunks"])
def test_llc_extra_metadata(llc_mds_datadirs, method):
    """Check that the LLC reads properly when using extra_metadata."""
    dirname, expected = llc_mds_datadirs
    nz, nface, ny, nx = expected['shape']
    nt = 1

    llc = xmitgcm.utils.get_extra_metadata(domain='llc', nx=nx)

    ds = xmitgcm.open_mdsdataset(dirname,
                                 iters=expected['test_iternum'],
                                 geometry=expected['geometry'],
                                 llc_method=method,
                                 extra_metadata=llc)

    assert ds.dims['face'] == 13
    assert ds.rA.dims == ('face', 'j', 'i')
    assert ds.rA.values.shape == (nface, ny, nx)
    assert ds.U.dims == ('time', 'k', 'face', 'j', 'i_g')
    assert ds.U.values.shape == (nt, nz, nface, ny, nx)
    assert ds.V.dims == ('time', 'k', 'face', 'j_g', 'i')
    assert ds.V.values.shape == (nt, nz, nface, ny, nx)

    if method == "smallchunks":
        assert ds.U.chunks == (nt*(1,), nz*(1,), nface*(1,), (ny,), (nx,))


def test_levels_diagnostics(mds_datadirs_with_inputfiles):
    dirname, expected = mds_datadirs_with_inputfiles

    for diagname, (levels, (idx, value)) in expected['diag_levels'].items():
        ds = xmitgcm.open_mdsdataset(dirname, prefix=[diagname], levels=levels,
                                     geometry=expected['geometry'])

        assert ds['Zl'].values[idx] == value

        with pytest.warns(UserWarning, match='nz will be ignored'):
            xmitgcm.open_mdsdataset(dirname, prefix=[diagname], levels=levels, 
                                    geometry=expected['geometry'], nz=12)


import cachetools.func
import os
import fnmatch

cache_maxsize = 100
cache_ttl = 600 # tem minutes

@cachetools.func.ttl_cache(maxsize=cache_maxsize, ttl=cache_ttl)
def listdir(path):
    return os.listdir(path)

@cachetools.func.ttl_cache(maxsize=cache_maxsize, ttl=cache_ttl)
def listdir_startswith(path, pattern):
    files = listdir(path)
    return [f for f in files if f.startswith(pattern)]

@cachetools.func.ttl_cache(maxsize=cache_maxsize, ttl=cache_ttl)
def listdir_endswith(path, pattern):
    files = listdir(path)
    return [f for f in files if f.endswith(pattern)]

@cachetools.func.ttl_cache(maxsize=cache_maxsize, ttl=cache_ttl)
def listdir_startsandendswith(path, start, end):
    files = listdir(path)
    return [f for f in files if f.endswith(end) and f.startswith(start)]

@cachetools.func.ttl_cache(maxsize=cache_maxsize, ttl=cache_ttl)
def listdir_fnmatch(path, pattern):
    files = listdir(path)
    return [f for f in files if fnmatch.fnmatch(f, pattern)]

def clear_cache():
    listdir.cache_clear()
    listdir_startswith.cache_clear()
    listdir_endswith.cache_clear()
    listdir_startsandendswith.cache_clear()
    listdir_fnmatch.cache_clear()



"""
All of the metadata related to MITgcm variables, grids, naming conventions, etc.
"""
# python 3 compatiblity
from __future__ import print_function, division
import numpy as np

# xarray>=0.12.0 compatiblity
try:
    from xarray.core.pycompat import OrderedDict
except ImportError:
    from collections import OrderedDict

# We are trying to combine the following two things:
# - MITgcm grid
#   http://mitgcm.org/sealion/online_documents/node47.html
# - Comodo conventions
#   http://pycomodo.forge.imag.fr/norm.html
# To do that, we have to set the proper metadata (attributes) for the MITgcm
# variables.

# The spatial dimensions, all 1D
# There is no "data" to go with these. They are just indices.
dimensions = OrderedDict(
    # x direction
    i=dict(dims=['i'], attrs=dict(
        standard_name="x_grid_index", axis='X',
        long_name="x-dimension of the t grid",
        swap_dim='XC')),
    i_g=dict(dims=['i_g'], attrs=dict(
        standard_name="x_grid_index_at_u_location", axis='X',
        long_name="x-dimension of the u grid", c_grid_axis_shift=-0.5,
        swap_dim='XG')),
    # i_z = dict(dims=['i_z'], swap_dim='XG', attrs=dict(
    #             standard_name="x_grid_index_at_f_location", axis='X',
    #             long_name="x-dimension of the f grid", c_grid_axis_shift=-0.5)),
    # y direction
    j=dict(dims=['j'], attrs=dict(
        standard_name="y_grid_index", axis='Y',
        long_name="y-dimension of the t grid", swap_dim='YC')),
    j_g=dict(dims=['j_g'], attrs=dict(
        standard_name="y_grid_index_at_v_location", axis='Y',
        long_name="y-dimension of the v grid", c_grid_axis_shift=-0.5,
        swap_dim='YG')),
    # j_z = dict(dims=['j_z'], swap_dim='YG', attrs=dict(
    #             standard_name="y_grid_index_at_f_location", axis='Y',
    #             long_name="y-dimension of the f grid", c_grid_axis_shift=-0.5)),
    # x direction
    k=dict(dims=['k'], attrs=dict(
        standard_name="z_grid_index", axis="Z",
        long_name="z-dimension of the t grid", swap_dim='Z')),
    k_u=dict(dims=['k_u'], attrs=dict(
        standard_name="z_grid_index_at_lower_w_location",
        axis="Z", long_name="z-dimension of the w grid",
        c_grid_axis_shift=0.5, swap_dim='Zu')),
    k_l=dict(dims=['k_l'], attrs=dict(
        standard_name="z_grid_index_at_upper_w_location",
        axis="Z", long_name="z-dimension of the w grid",
        c_grid_axis_shift=-0.5, swap_dim='Zl')),
    # this is complicated because it is offset in both directions - allowed by comodo?
    k_p1=dict(dims=['k_p1'], attrs=dict(
        standard_name="z_grid_index_at_w_location",
        axis="Z", long_name="z-dimension of the w grid",
        c_grid_axis_shift=(-0.5, 0.5), swap_dim='Zp1'))
)

horizontal_coordinates_spherical = OrderedDict(
    XC=dict(dims=["j", "i"], attrs=dict(
        standard_name="longitude", long_name="longitude",
        units="degrees_east", coordinate="YC XC")),
    YC=dict(dims=["j", "i"], attrs=dict(
        standard_name="latitude", long_name="latitude",
        units="degrees_north", coordinate="YC XC")),
    XG=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="longitude_at_f_location", long_name="longitude",
        units="degrees_east", coordinate="YG XG")),
    YG=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="latitude_at_f_location", long_name="latitude",
        units="degrees_north", coordinate="YG XG"))
)

horizontal_coordinates_llc = horizontal_coordinates_spherical.copy()
horizontal_coordinates_llc.update(OrderedDict(
    CS=dict(dims=["j", "i"], attrs=dict(standard_name="Cos of grid orientation angle",
                                        long_name="AngleCS", units=" ", coordinate="YC XC"),
            filename='AngleCS'),
    SN=dict(dims=["j", "i"], attrs=dict(standard_name="Sin of grid orientation angle",
                                        long_name="AngleSN", units=" ", coordinate="YC XC"),
            filename='AngleSN')
)
)

horizontal_coordinates_curvcart = OrderedDict(
    XC=dict(dims=["j", "i"], attrs=dict(
        standard_name="plane_x_coordinate", long_name="x coordinate",
        units="m", coordinate="YC XC")),
    YC=dict(dims=["j", "i"], attrs=dict(
        standard_name="plane_y_coordinate", long_name="y coordinate",
        units="m", coordinate="YC XC")),
    XG=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="plane_x_coordinate_at_f_location",
        long_name="x coordinate", units="m", coordinate="YG XG")),
    YG=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="plane_y_coordinate_at_f_location",
        long_name="y coordinate", units="m", coordinate="YG XG")),
    CS=dict(dims=["j", "i"], attrs=dict(standard_name="Cos of grid orientation angle",
                                        long_name="AngleCS", units=" ", coordinate="YC XC"),
            filename='AngleCS'),
    SN=dict(dims=["j", "i"], attrs=dict(standard_name="Sin of grid orientation angle",
                                        long_name="AngleSN", units=" ", coordinate="YC XC"),
            filename='AngleSN')
)

horizontal_coordinates_cartesian = OrderedDict(
    XC=dict(dims=["j", "i"], attrs=dict(
        standard_name="plane_x_coordinate", long_name="x coordinate",
        units="m", coordinate="YC XC")),
    YC=dict(dims=["j", "i"], attrs=dict(
        standard_name="plane_y_coordinate", long_name="y coordinate",
        units="m", coordinate="YC XC")),
    XG=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="plane_x_coordinate_at_f_location",
        long_name="x coordinate", units="m", coordinate="YG XG")),
    YG=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="plane_y_coordinate_at_f_location",
        long_name="y coordinate", units="m", coordinate="YG XG"))
)

vertical_coordinates = OrderedDict(
    Z=dict(dims=["k"], attrs=dict(
        standard_name="depth",
        long_name="vertical coordinate of cell center",
        units="m", positive="down"),
        filename="RC", slice=(slice(None), 0, 0)),
    Zp1=dict(dims=["k_p1"], attrs=dict(
        standard_name="depth_at_w_location",
        long_name="vertical coordinate of cell interface",
        units="m", positive="down"),
        filename="RF", slice=(slice(None), 0, 0)),
    Zu=dict(dims=["k_u"], attrs=dict(
        standard_name="depth_at_lower_w_location",
        long_name="vertical coordinate of lower cell interface",
        units="m", positive="down"),
        filename="RF", slice=(slice(1, None), 0, 0)),
    Zl=dict(dims=["k_l"], attrs=dict(
        standard_name="depth_at_upper_w_location",
        long_name="vertical coordinate of upper cell interface",
        units="m", positive="down"),
        filename="RF", slice=(slice(None, -1), 0, 0))
)

# I got these variable names from the default MITgcm netcdf output
horizontal_grid_variables = OrderedDict(
    # tracer cell
    rA=dict(dims=["j", "i"], attrs=dict(standard_name="cell_area",
                                        long_name="cell area", units="m2", coordinate="YC XC"),
            filename='RAC'),
    dxG=dict(dims=["j_g", "i"], attrs=dict(
        standard_name="cell_x_size_at_v_location",
        long_name="cell x size", units="m", coordinate="YG XC"),
        filename='DXG'),
    dyG=dict(dims=["j", "i_g"], attrs=dict(
        standard_name="cell_y_size_at_u_location",
        long_name="cell y size", units="m", coordinate="YC XG"),
        filename='DYG'),
    Depth=dict(dims=["j", "i"], attrs=dict(standard_name="ocean_depth",
                                           long_name="ocean depth", units="m", coordinate="XC YC")),
    # vorticity cell
    rAz=dict(dims=["j_g", "i_g"], attrs=dict(
        standard_name="cell_area_at_f_location",
        long_name="cell area", units="m", coordinate="YG XG"),
        filename='RAZ'),
    dxC=dict(dims=["j", "i_g"], attrs=dict(
        standard_name="cell_x_size_at_u_location",
        long_name="cell x size", units="m", coordinate="YC XG"),
        filename='DXC'),
    dyC=dict(dims=["j_g", "i"], attrs=dict(
        standard_name="cell_y_size_at_v_location",
        long_name="cell y size", units="m", coordinate="YG XC"),
        filename='DYC'),
    # u cell
    rAw=dict(dims=["j", "i_g"], attrs=dict(
        standard_name="cell_area_at_u_location",
        long_name="cell area", units="m2", coordinate="YG XC"),
        filename='RAW'),
    # v cell
    rAs=dict(dims=["j_g", "i"], attrs=dict(
        standard_name="cell_area_at_v_location",
        long_name="cell area", units="m2", coordinate="YG XC"),
        filename='RAS'),
)

vertical_grid_variables = OrderedDict(
    drC=dict(dims=['k_p1'], attrs=dict(
        standard_name="cell_z_size_at_w_location",
        long_name="cell z size", units="m"), filename='DRC'),
    drF=dict(dims=['k'], attrs=dict(
        standard_name="cell_z_size",
        long_name="cell z size", units="m"), filename='DRF'),
    PHrefC=dict(dims=['k'], attrs=dict(
        standard_name="cell_reference_pressure",
        long_name='Reference Hydrostatic Pressure', units='m2 s-2')),
    PHrefF=dict(dims=['k_p1'], attrs=dict(
        standard_name="cell_reference_pressure",
        long_name='Reference Hydrostatic Pressure', units='m2 s-2'))
)

volume_grid_variables = OrderedDict(
    hFacC=dict(dims=['k', 'j', 'i'], attrs=dict(
        standard_name="cell_vertical_fraction",
        long_name="vertical fraction of open cell")),
    hFacW=dict(dims=['k', 'j', 'i_g'], attrs=dict(
        standard_name="cell_vertical_fraction_at_u_location",
        long_name="vertical fraction of open cell")),
    hFacS=dict(dims=['k', 'j_g', 'i'], attrs=dict(
        standard_name="cell_vertical_fraction_at_v_location",
        long_name="vertical fraction of open cell"))
)

# Mask files denoting wet points
# set filename to hFac then compute mask as 1 where hFacC nonzero
mask_variables = OrderedDict(
    maskC=dict(dims=['k', 'j', 'i'], attrs=dict(
        standard_name="sea_binary_mask_at_t_location",
        long_name="mask denoting wet point at center"),
        filename="hFacC"),
    maskW=dict(dims=['k', 'j', 'i_g'], attrs=dict(
        standard_name="cell_vertical_fraction_at_u_location",
        long_name="mask denoting wet point at interface"),
        filename="hFacW"),
    maskS=dict(dims=['k', 'j_g', 'i'], attrs=dict(
        standard_name="cell_vertical_fraction_at_v_location",
        long_name="mask denoting wet point at interface"),
        filename="hFacS")
)

# this a template: NAME gets replaced with the layer name (e.g. 1RHO)
#                  dim gets prepended with the layer number (e.g. l1_b)
layers_grid_variables = OrderedDict(
    layer_NAME_bounds=dict(dims=['l_b'], attrs=dict(
        standard_name="ocean_layer_coordinate_NAME_bounds",
        long_name="boundaries points of layer NAME",
        axis="NAME", c_grid_axis_shift=-0.5),
        filename="layersNAME", slice=(slice(None), 0, 0)),
    layer_NAME_center=dict(dims=['l_c'], attrs=dict(
        standard_name="ocean_layer_coordinate_NAME_center",
        long_name="center points of layer NAME",
        axis="NAME"),
        filename="layersNAME", slice=(slice(None), 0, 0),
        # if we don't convert to array, dask can't tokenize
        # https://github.com/pydata/xarray/issues/1014
        transform=(lambda x: np.asarray(0.5*(x[1:] + x[:-1])))),
    layer_NAME_interface=dict(dims=['l_i'], attrs=dict(
        standard_name="ocean_layer_coordinate_NAME_interface",
        long_name="interface points of layer NAME",
        axis="NAME", c_grid_axis_shift=-0.5),
        filename="layersNAME", slice=(slice(1, -1), 0, 0))
)

# _grid_special_mapping = {
# # name: (file_name, slice_to_extract, expecting_3D_field)
#     'Z': ('RC', (slice(None),0,0), 3),
#     'Zp1': ('RF', (slice(None),0,0), 3),
#     'Zu': ('RF', (slice(1,None),0,0), 3),
#     'Zl': ('RF', (slice(None,-1),0,0), 3),
#     # this will create problems with some curvillinear grids
#     # whate if X and Y need to be 2D?
#     'X': ('XC', (0,slice(None)), 2),
#     'Y': ('YC', (slice(None),0), 2),
#     'Xp1': ('XG', (0,slice(None)), 2),
#     'Yp1': ('YG', (slice(None),0), 2),
#     'rA': ('RAC', (slice(None), slice(None)), 2),
#     'HFacC': ('hFacC', 3*(slice(None),), 3),
#     'HFacW': ('hFacW', 3*(slice(None),), 3),
#     'HFacS': ('hFacS', 3*(slice(None),), 3),
# }


# also try to match CF standard names
# http://cfconventions.org/Data/cf-standard-names/28/build/cf-standard-name-table.html

state_variables = OrderedDict(
    # default state variables
    U = dict(dims=['k','j','i_g'], attrs=dict(
                standard_name='sea_water_x_velocity', mate='V',
                long_name='Zonal Component of Velocity', units='m s-1')),
    V = dict(dims=['k','j_g','i'], attrs=dict(
                standard_name='sea_water_y_velocity', mate='U',
                long_name='Meridional Component of Velocity', units='m s-1')),
    W = dict(dims=['k_l','j','i'], attrs=dict(
                standard_name='sea_water_z_velocity',
                long_name='Vertical Component of Velocity', units='m s-1')),
    T = dict(dims=['k','j','i'], attrs=dict(
                standard_name="sea_water_potential_temperature",
                long_name='Potential Temperature', units='degree_Celcius')),
    S = dict(dims=['k','j','i'], attrs=dict(
                standard_name="sea_water_salinity",
                long_name='Salinity', units='psu')),
    PH= dict(dims=['k','j','i'], attrs=dict(
                standard_name="sea_water_dynamic_pressue",
                long_name='Hydrostatic Pressure Pot.(p/rho) Anomaly',
                units='m2 s-2')),
    PHL=dict(dims=['j','i'], attrs=dict(
                standard_name="sea_water_dynamic_pressure_at_sea_floor",
                long_name='Bottom Pressure Pot.(p/rho) Anomaly',
                units='m2 s-2')),
    Eta=dict(dims=['j','i'], attrs=dict(
                standard_name="sea_surface_height_above_geoid",
                long_name='Surface Height Anomaly', units='m')),
    # default tave variables
    uVeltave = dict(dims=['k','j','i_g'], attrs=dict(
                standard_name='sea_water_x_velocity', mate='vVeltave',
                long_name='Zonal Component of Velocity', units='m s-1')),
    vVeltave = dict(dims=['k','j_g','i'], attrs=dict(
                standard_name='sea_water_y_velocity', mate='uVeltave',
                long_name='Meridional Component of Velocity', units='m s-1')),
    wVeltave = dict(dims=['k_l','j','i'], attrs=dict(
                standard_name='sea_water_z_velocity',
                long_name='Vertical Component of Velocity', units='m s-1')),
    Ttave = dict(dims=['k','j','i'], attrs=dict(
                standard_name="sea_water_potential_temperature",
                long_name='Potential Temperature', units='degree_Celcius')),
    Stave = dict(dims=['k','j','i'], attrs=dict(
                standard_name="sea_water_salinity",
                long_name='Salinity', units='psu')),
    PhHytave= dict(dims=['k','j','i'], attrs=dict(
                standard_name="sea_water_dynamic_pressue",
                long_name='Hydrostatic Pressure Pot.(p/rho) Anomaly',
                units='m2 s-2')),
    PHLtave=dict(dims=['j','i'], attrs=dict(
                standard_name="sea_water_dynamic_pressure_at_sea_floor",
                long_name='Bottom Pressure Pot.(p/rho) Anomaly',
                units='m2 s-2')),
    ETAtave=dict(dims=['j','i'], attrs=dict(
                standard_name="sea_surface_height_above_geoid",
                long_name='Surface Height Anomaly', units='m')),
    # TODO: finish encoding this crap!
    Convtave=dict(dims=['k_l','j','i'], attrs=dict(
                standard_name="convective_adjustment_index",
                long_name="Convective Adjustment Index")),
    Eta2tave=dict(dims=['j','i'], attrs=dict(
                standard_name="square_of_sea_surface_height_above_geoid",
                long_name='Square of Surface Height Anomaly', units='m2')),
    PHL2tave=dict(dims=['j','i'], attrs=dict(
                standard_name="square_of_sea_water_dynamic_pressure_at_sea_floor",
                long_name='Square of Bottom Pressure Pot.(p/rho) Anomaly',
                units='m4 s-4')),
    sFluxtave=dict(dims=['j','i'], attrs=dict(
                standard_name="virtual_salt_flux_into_sea_water",
                long_name='total salt flux (match salt-content variations), '
                          '>0 increases salt', units='g m-2 s-1')),
    Tdiftave=dict(dims=['k_l','j','i'], attrs=dict(
                standard_name="upward_potential_temperature_flux_"
                              "due_to_diffusion",
                long_name="Vertical Diffusive Flux of Pot.Temperature",
                units="K m3 s-1")),
    tFluxtave=dict(dims=['j','i'], attrs=dict(
                standard_name="surface_downward_heat_flux_in_sea_water",
                long_name="Total heat flux (match heat-content variations), "
                          ">0 increases theta", units="W m-2")),
    TTtave = dict(dims=['k','j','i'], attrs=dict(
                standard_name="square_of_sea_water_potential_temperature",
                long_name='Square of Potential Temperature',
                units='degree_Celcius')),
    uFluxtave=dict(dims=['j','i_g'], attrs=dict(
                standard_name="surface_downward_x_stress",
                long_name='surface zonal momentum flux, positive -> increase u',
                units='N m-2', mate='vFluxtave')),
    UStave=dict(dims=['k','j','i_g'], attrs=dict(
                standard_name="product_of_sea_water_x_velocity_and_salinity",
                long_name="Zonal Transport of Salinity",
                units="psu m s-1", mate='VStave')),
    UTtave=dict(dims=['k','j','i_g'], attrs=dict(
                standard_name="product_of_sea_water_x_velocity_and_"
                              "potential_temperature",
                long_name="Zonal Transport of Potential Temperature",
                units="K m s-1", mate='VTtave')),
    UUtave = dict(dims=['k','j','i_g'], attrs=dict(
                standard_name='square_of_sea_water_x_velocity',
                long_name='Square of Zonal Component of Velocity',
                units='m2 s-2')),
    UVtave=dict(dims=['k','j_g','i_g'], attrs=dict(
                standard_name="product_of_sea_water_x_velocity_and_"
                              "sea_water_y_velocity",
                long_name="Product of meridional and zonal velocity",
                units="m2 s-2")),
    vFluxtave=dict(dims=['j_g','i'], attrs=dict(
                standard_name="surface_downward_y_stress",
                long_name='surface meridional momentum flux, '
                          'positive -> increase u',
                units='N m-2', mate='uFluxtave')),
    VStave=dict(dims=['k','j_g','i'], attrs=dict(
                standard_name="product_of_sea_water_y_velocity_and_salinity",
                long_name="Meridional Transport of Salinity",
                units="psu m s-1", mate='UStave')),
    VTtave=dict(dims=['k','j_g','i'], attrs=dict(
                standard_name="product_of_sea_water_y_velocity_and_"
                              "potential_temperature",
                long_name="Meridional Transport of Potential Temperature",
                units="K m s-1", mate='UTtave')),
    VVtave = dict(dims=['k','j_g','i'], attrs=dict(
                standard_name='square_of_sea_water_y_velocity',
                long_name='Square of Meridional Component of Velocity',
                units='m2 s-2')),
    WStave=dict(dims=['k_l','j','i'], attrs=dict(
                standard_name="product_of_sea_water_z_velocity_and_salinity",
                long_name="Vertical Transport of Salinity",
                units="psu m s-1")),
    WTtave=dict(dims=['k_l','j','i'], attrs=dict(
                standard_name="product_of_sea_water_z_velocity_and_"
                              "potential_temperature",
                long_name="Vertical Transport of Potential Temperature",
                units="K m s-1")),
)

for n in range(1,100):
   trname = 'PTRACER%02d' % n
   state_variables[trname] = dict(dims=['k','j','i'],
       attrs=dict(standard_name='%s_concentration' % trname,
                  long_name="Concentration of %s" % trname,
                  units="kg m-3")
   )
for n in range(1,100):
   travename = 'PTRtave%02d' % n
   state_variables[travename] = dict(dims=['k','j','i'],
       attrs=dict(standard_name='%s_concentration' % travename,
                  long_name="Concentration of %s" % travename,
                  units="kg m-3")
   )


# these aliases are necessary to deal with the LLC output which was saved with
# unconventional variable names and no meta files
aliases = {'Eta': 'ETAN', 'PhiBot': 'PHIBOT', 'Salt': 'SALT', 'Theta': 'THETA'}

package_state_variables = {
    'KPPviscAz': dict(dims=['k_l', 'j', 'i'], attrs=dict(
        standard_name="KPP_vertical_eddy_visc",
        long_name='KPP vertical eddy viscosity coefficient',
        units='m2 s-1')),
    'KPPdiffKzS': dict(dims=['k_l', 'j', 'i'], attrs=dict(
        standard_name="KPP_diff_tracer_salt",
        long_name='KPP vertical diffusion coefficient for salt and tracers',
        units='m2 s-1')),
    'KPPdiffKzT': dict(dims=['k_l', 'j', 'i'], attrs=dict(
        standard_name="KPP_diff_temp",
        long_name='KPP vertical diffusion coefficient for heat',
        units='m2 s-1')),
    'KPPghat': dict(dims=['k_l', 'j', 'i'], attrs=dict(
        standard_name="KPP_non_local_coeff",
        long_name='non-local transfer coefficient',
        units='s m-2 ')),
    'KPPhbl': dict(dims=['j', 'i'], attrs=dict(
        standard_name="KPP_boundary_layer",
        long_name='KPP boundary layer depth, bulk Ri criterion',
        units='m')),
    'KPPg_TH': dict(dims=['k_l','j','i'], attrs=dict(
        standard_name='KPP_theta_flux',
        long_name='KPP non-local Flux of Pot.Temperature',
        units='degC m3 s-1')),
    'KPPg_SLT': dict(dims=['k_l','j','i'], attrs=dict(
        standard_name='KPP_salt_flux',
        long_name='KPP non-local Flux of Salinity',
        units='psu m3 s-1')),
    # pkg/thsice variables
    'ice_fract': dict(dims=['j', 'i'], attrs=dict(
        standard_name="sea_ice_area_fraction",
        long_name='THSICE sea fractional ice area',
        units=' ')),
    'ice_iceH': dict(dims=['j', 'i'], attrs=dict(
        standard_name="sea_ice_thickness",
        long_name='THSICE actual sea ice thickness',
        units='m')),
    'ice_snowH': dict(dims=['j', 'i'], attrs=dict(
        standard_name="snow_thickness",
        long_name='THSICE actual snow thickness',
        units='m')),
    'ice_snowAge': dict(dims=['j', 'i'], attrs=dict(
        standard_name="snow_age",
        long_name='THSICE snow age',
        units='s')),
    'ice_Tsrf': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_temperature",
        long_name='temperature at surface',
        units='degC')),
    'ice_Tice1': dict(dims=['j', 'i'], attrs=dict(
        standard_name="temperature_of_sea_ice_layer1",
        long_name='THSICE temperature of ice layer 1',
        units='degC')),
    'ice_Tice2': dict(dims=['j', 'i'], attrs=dict(
        standard_name="temperature_of_sea_ice_layer2",
        long_name='THSICE temperature of ice layer 2',
        units='degC')),
    'ice_Qice1': dict(dims=['j', 'i'], attrs=dict(
        standard_name="enthalpy_content_of_sea_ice_layer1",
        long_name='THSICE enthalpy of ice layer 1',
        units='J kg-1')),
    'ice_Qice2': dict(dims=['j', 'i'], attrs=dict(
        standard_name="enthalpy_content_of_sea_ice_layer2",
        long_name='THSICE enthalpy of ice layer 2',
        units='J kg-1')),
    'ice_flxAtm': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_downward_heat_flux",
        long_name='THSICE heat flux over ice from the atmosphere (+=down)',
        units='W m-2')),
    'ice_frwAtm': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_upward_freshwater_flux",
        long_name='freshwater flux (E-P) over ice to the atmosphere (+=up)',
        units='kg m-2 s-1')),
    'ice_tOceMxL': dict(dims=['j', 'i'], attrs=dict(
        standard_name="temperature_of_ocean_mixed_layer",
        long_name='temperature of ocean mixed layer',
        units='degC')),
    'ice_sOceMxL': dict(dims=['j', 'i'], attrs=dict(
        standard_name="salinity_of_ocean_mixed_layer",
        long_name='salinity of ocean mixed layer',
        units=' ')),
    # pkg/seaice variables
    'AREA': dict(dims=['j', 'i'], attrs=dict(
        standard_name="sea_ice_area_fraction",
        long_name='SEAICE sea fractional ice area',
        units=' ')),
    'HEFF': dict(dims=['j', 'i'], attrs=dict(
        standard_name="sea_ice_thickness",
        long_name='SEAICE mean sea ice thickness per grid cell',
        units='m')),
    'HSNOW': dict(dims=['j', 'i'], attrs=dict(
        standard_name="snow_thickness",
        long_name='SEAICE mean snow ice thickness per grid cell',
        units='m')),
    'UICE': dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="sea_ice_x_velocity", mate='VICE',
        long_name='SEAICE ice drift velocity in x/i-direction',
        units='m s-1')),
    'VICE': dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="sea_ice_y_velocity", mate='UICE',
        long_name='SEAICE ice drift velocity in y/j-direction',
        units='m s-1')),
    'Qnet': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_upward_heat_flux",
        long_name='SEAICE net upward heat flux (+=up)',
        units='W m-2')),
    'Qsw': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_upward_shortwave_flux",
        long_name='SEAICE upward shortwave heat flux (+=up)',
        units='W m-2')),
    'EmPmR': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_upward_freshwater_flux",
        long_name='SEAICE upward freshwater flux (+=up)',
        units='kg m-2 s-1')),
    'UWIND': dict(dims=['j', 'i'], attrs=dict(
        standard_name="x_wind", mate='VWIND',
        long_name='SEAICE wind forcing in x/i-direction',
        units='m s-1')),
    'VWIND': dict(dims=['j', 'i'], attrs=dict(
        standard_name="y_wind", mate='UWIND',
        long_name='SEAICE wind forcing in y/j-direction',
        units='m s-1')),
    'FU': dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="downward_eastward_stress_at_sea_ice_base",
        mate='FV',
        long_name='SEAICE surface stress on ocean in x/i-direction',
        units='N m-2')),
    'FV': dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="downward_northward_stress_at_sea_ice_base",
        mate='FU',
        long_name='SEAICE surface stress on ocean in y/j-direction',
        units='N m-2')),
    'AREAtave': dict(dims=['j', 'i'], attrs=dict(
        standard_name="sea_ice_area_fraction",
        long_name='SEAICE time averaged sea fractional ice area',
        units=' ')),
    'HEFFtave': dict(dims=['j', 'i'], attrs=dict(
        standard_name="sea_ice_thickness",
        long_name='SEAICE mean sea ice thickness per grid cell',
        units='m')),
    'UICEtave': dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="sea_ice_x_velocity", mate='VICEtave',
        long_name='SEAICE ice drift velocity in x/i-direction',
        units='m s-1')),
    'VICEtave': dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="sea_ice_y_velocity", mate='UICEtave',
        long_name='SEAICE ice drift velocity in y/j-direction',
        units='m s-1')),
    'FUtave': dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="downward_eastward_stress_at_sea_ice_base",
        mate='FVtave',
        long_name='SEAICE surface stress on ocean in x/i-direction',
        units='N m-2')),
    'FVtave': dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="downward_northward_stress_at_sea_ice_base",
        mate='FUtave',
        long_name='SEAICE surface stress on ocean in y/j-direction',
        units='N m-2')),
    'EmPmRtave': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_upward_freshwater_flux",
        long_name='SEAICE upward freshwater flux (+=up)',
        units='kg m-2 s-1')),
    'QNETtave': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_upward_shortwave_flux",
        long_name='SEAICE net upward heat flux (+=up)',
        units='W m-2')),
    'QSWtave': dict(dims=['j', 'i'], attrs=dict(
        standard_name="surface_net_upward_shortwave_flux",
        long_name='SEAICE upward freshwater flux (+=up)',
        units='kg m-2 s-1'))
}

extra_grid_variables = OrderedDict(
    # Printed from write_grid when debugLevel>=debugLevC
    rLowC=dict(dims=['j', 'i'], attrs=dict(
        standard_name="depth_r0_to_bottom",
        long_name='Depth fixed r0 to bottom at tracer location',
        units='m')),
    rLowW=dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="depth_r0_to_bottom_at_u_location",
        long_name='Depth from fixed r0 to bottom at u location',
        units='m')),
    rLowS=dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="depth_r0_to_bottom_at_v_location",
        long_name='Depth fixed r0 to bottom at v location',
        units='m')),
    rSurfC=dict(dims=['j', 'i'], attrs=dict(
        standard_name="depth_r0_to_ref_surface",
        long_name='Depth fixed r0 to reference surface level at '
                  'tracer location',
        units='m')),
    rSurfW=dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="depth_r0_to_ref_surface_at_u_location",
        long_name='Depth fixed r0 to reference surface level at u location',
        units='m')),
    rSurfS=dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="depth_r0_to_ref_surface_at_v_location",
        long_name='Depth fixed r0 to reference surface level at v location',
        units='m')),
    # Printed from write_grid when useOBCS==T
    maskInC=dict(dims=['j', 'i'], attrs=dict(
        standard_name="interior_2d_mask",
        long_name='OBCS 2D interior mask at tracer location, zero beyond OB',
        units='')),
    maskInW=dict(dims=['j', 'i_g'], attrs=dict(
        standard_name="interior_2d_mask_at_u_location",
        long_name='OBCS 2D interior mask at u location, zero on & beyond OB',
        units='')),
    maskInS=dict(dims=['j_g', 'i'], attrs=dict(
        standard_name="interior_2d_mask_at_v_location",
        long_name='OBCS 2D interior mask at v location, zero on & beyond OB',
        units='')),
    # Printed from pkg/ctrl/ctrl_init_wet when useCTRL=T
    maskCtrlC=dict(dims=['k', 'j', 'i'], attrs=dict(
        standard_name="ctrl_vector_3d_mask",
        long_name='CTRL 3D mask where ctrl vector is active at '
                  'tracer location',
        units='')),
    maskCtrlW=dict(dims=['k', 'j', 'i_g'], attrs=dict(
        standard_name="ctrl_vector_3d_mask_at_u_location",
        long_name='CTRL 3D mask where ctrl vector is active at u location',
        units='')),
    maskCtrlS=dict(dims=['k', 'j_g', 'i'], attrs=dict(
        standard_name="ctrl_vector_3d_mask_at_v_location",
        long_name='CTRL 3D mask where ctrl vector is active at v location',
        units='')),
    # Reference density profile
    rhoRef=dict(dims=['k'],attrs=dict(
        standard_name="reference_density_profile",
        long_name="1D, vertical reference density profile",
        coordinate="Z",
        units='kg m-3'),
        filename='RhoRef'),
    # Additional grid metrics
    dxF=dict(dims=['j','i'],attrs=dict(
        standard_name="cell_x_size_at_t_location",
        long_name="cell x size", units="m", coordinate="YC XC"),
        filename='DXF'),
    dyF=dict(dims=['j','i'],attrs=dict(
        standard_name="cell_y_size_at_t_location",
        long_name="cell y size", units="m", coordinate="YC XC"),
        filename='DYF'),
    dxV=dict(dims=['j_g','i_g'],attrs=dict(
        standard_name="cell_x_size_at_f_location",
        long_name="cell x size", units="m", coordinate="YG XG"),
        filename='DXV'),
    dyU=dict(dims=['j_g','i_g'],attrs=dict(
        standard_name="cell_y_size_at_f_location",
        long_name="cell y size", units="m", coordinate="YG XG"),
        filename='DYU')
    # Printed from write_grid when sigma coordinates are used
    # AHybSigmF, BHybSigF, ...
    # Unclear where on the grid these variables exist,
    # skipping for now ...
)

# Nptracers=99
# _ptracers = { 'PTRACER%02d' % n :
#                (dims=['k','j','i'], 'PTRACER%02d Concentration' % n, "tracer units/m^3")
#                for n in range(Nptracers)}

# these variable names have hyphens and need a different syntax
# state_variables['GM_Kwx-T'] = dict(
#     dims=['k_l','j','i'], attrs=dict(
#         standard_name="K_31_element_of_GMRedi_tensor",
#         long_name="K_31 element (W.point, X.dir) of GM-Redi tensor",
#         units="m2 s-1"
#     )
# )
#
# state_variables['GM_Kwy-T'] = dict(
#     dims=['k_l','j','i'], attrs=dict(
#         standard_name="K_32_element_of_GMRedi_tensor",
#         long_name="K_32 element (W.point, Y.dir) of GM-Redi tensor",
#         units="m2 s-1"
#     )
# )
#
# state_variables['GM_Kwy-T'] = dict(
#     dims=['k_l','j','i'], attrs=dict(
#         standard_name="K_33_element_of_GMRedi_tensor",
#         long_name="K_33 element (W.point, Z.dir) of GM-Redi tensor",
#         units="m2 s-1"
#     )
# )


# Nptracers=99
# _ptracers = { 'PTRACER%02d' % n :
#                (dims=['k','j','i'], 'PTRACER%02d Concentration' % n, "tracer units/m^3")
#                for n in range(Nptracers)}



import pytest

from xmitgcm import file_utils

@pytest.fixture(scope="session")
def directory_with_files(tmpdir_factory):
    temppath = tmpdir_factory.mktemp("xmitgcm_test_data")
    temppath.join('bar.0000000001.meta').ensure(file=True)
    temppath.join('baz.data').ensure(file=True)
    return temppath

def test_listdir(directory_with_files):
    path = str(directory_with_files)
    assert sorted(file_utils.listdir(path)) == sorted(['bar.0000000001.meta', 'baz.data'])

def test_listdir_startswith(directory_with_files):
    path = str(directory_with_files)
    assert file_utils.listdir_startswith(path, 'bar') == ['bar.0000000001.meta']

def test_listdir_endswith(directory_with_files):
    path = str(directory_with_files)
    assert file_utils.listdir_endswith(path, '.data') == ['baz.data']

def test_listdir_startsandendswith(directory_with_files):
    path = str(directory_with_files)
    assert file_utils.listdir_startsandendswith(path, 'bar', '.meta') == ['bar.0000000001.meta']

def test_listdir_fnmatch(directory_with_files):
    path = str(directory_with_files)
    assert file_utils.listdir_fnmatch(path, '*.??????????.meta') == ['bar.0000000001.meta']

def test_clear_cache():
    file_utils.clear_cache()



import pytest
import os
import numpy as np
import xarray
import dask
from xmitgcm.test.test_xmitgcm_common import (hide_file, file_md5_checksum,
    all_mds_datadirs, mds_datadirs_with_diagnostics, llc_mds_datadirs,
    layers_mds_datadirs, all_grid_datadirs, mds_datadirs_with_inputfiles,
    _experiments)
from xmitgcm.file_utils import listdir


_xc_meta_content = """ simulation = { 'global_oce_latlon' };
 nDims = [   2 ];
 dimList = [
    90,    1,   90,
    40,    1,   40
 ];
 dataprec = [ 'float32' ];
 nrecords = [     1 ];
"""


def test_parse_meta(tmpdir):
    """Check the parsing of MITgcm .meta into python dictionary."""

    from xmitgcm.utils import parse_meta_file
    p = tmpdir.join("XC.meta")
    p.write(_xc_meta_content)
    fname = str(p)
    result = parse_meta_file(fname)
    expected = {
        'nrecords': 1,
        'basename': 'XC',
        'simulation': "'global_oce_latlon'",
        'dimList': [[90, 1, 90], [40, 1, 40]],
        'nDims': 2,
        'dataprec': np.dtype('float32')
    }
    for k, v in expected.items():
        assert result[k] == v


@pytest.mark.parametrize("dtype", [np.dtype('f8'), np.dtype('f4'), np.dtype('i4')])
def test_read_raw_data(tmpdir, dtype):
    """Check our utility for reading raw data."""

    from xmitgcm.utils import read_raw_data
    shape = (2, 4)
    # create some test data
    testdata = np.zeros(shape, dtype)
    # write to a file
    datafile = tmpdir.join("tmp.data")
    datafile.write_binary(testdata.tobytes())
    fname = str(datafile)
    # now test the function
    data = read_raw_data(fname, dtype, shape)
    np.testing.assert_allclose(data, testdata)
    # interestingly, memmaps are also ndarrays, but not vice versa
    assert isinstance(data, np.ndarray) and not isinstance(data, np.memmap)
    # check memmap
    mdata = read_raw_data(fname, dtype, shape, use_mmap=True)
    assert isinstance(mdata, np.memmap)

    # make sure errors are correct
    wrongshape = (2, 5)
    with pytest.raises(IOError):
        read_raw_data(fname, dtype, wrongshape)

    # test optional functionalities
    shape = (5, 15, 10)
    shape_subset = (15, 10)
    testdata = np.zeros(shape, dtype)
    # create some test data
    x = np.arange(shape[0], dtype=dtype)
    for k in np.arange(shape[0]):
        testdata[k, :, :] = x[k]
    # write to a file
    datafile = tmpdir.join("tmp.data")
    datafile.write_binary(testdata.tobytes())
    fname = str(datafile)
    # now test the function
    for k in np.arange(shape[0]):
        offset = (k * shape[1] * shape[2] * dtype.itemsize)
        data = read_raw_data(fname, dtype, shape_subset,
                             offset=offset, partial_read=True)
        np.testing.assert_allclose(data, testdata[k, :, :])
        assert isinstance(data, np.ndarray) and not isinstance(
            data, np.memmap)
        # check memmap
        mdata = read_raw_data(fname, dtype, shape_subset,
                              offset=offset, partial_read=True,
                              use_mmap=True)
        assert isinstance(mdata, np.memmap)

        # test it breaks when it should
        with pytest.raises(IOError):
            # read with wrong shape
            read_raw_data(fname, dtype, shape_subset,
                          offset=0, partial_read=False)
        with pytest.raises(IOError):
            read_raw_data(fname, dtype, shape_subset,
                          offset=0, partial_read=False, use_mmap=True)
        with pytest.raises(ValueError):
            # use offset when trying to read global file
            read_raw_data(fname, dtype, shape_subset,
                          offset=4, partial_read=False)
        with pytest.raises(ValueError):
            read_raw_data(fname, dtype, shape_subset,
                          offset=4, partial_read=False, use_mmap=True)
            # offset is too big
        with pytest.raises(ValueError):
            read_raw_data(fname, dtype, shape, offset=(
                shape[0]*shape[1]*shape[2]*dtype.itemsize), partial_read=True)
        with pytest.raises(ValueError):
            read_raw_data(fname, dtype, shape, offset=(
                shape[0]*shape[1]*shape[2]*dtype.itemsize), partial_read=True,
                use_mmap=True)

# a meta test of our own utitity funcion
def test_file_hiding(all_mds_datadirs):
    dirname, _ = all_mds_datadirs
    basenames = ['XC.data', 'XC.meta']
    listed_files = listdir(dirname)
    for basename in basenames:
        assert os.path.exists(os.path.join(dirname, basename))
        assert basename in listed_files
    with hide_file(dirname, *basenames):
        listed_files = listdir(dirname)
        for basename in basenames:
            assert not os.path.exists(os.path.join(dirname, basename))
            assert basename not in listed_files
    for basename in basenames:
        assert os.path.exists(os.path.join(dirname, basename))


def test_read_mds(all_mds_datadirs):
    """Check that we can read mds data from .meta / .data pairs"""

    dirname, expected = all_mds_datadirs

    from xmitgcm.utils import read_mds

    prefix = 'XC'
    basename = os.path.join(dirname, prefix)
    # should be dask by default
    res = read_mds(basename)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], dask.array.core.Array)

    # try some options
    res = read_mds(basename, use_dask=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.memmap)

    res = read_mds(basename, use_dask=False, use_mmap=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.ndarray)

    res = read_mds(basename, chunks="2D")
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], dask.array.core.Array)

    res = read_mds(basename, chunks="2D", use_dask=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.memmap)

    res = read_mds(basename, chunks="2D", use_dask=False,
                   use_mmap=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.ndarray)

    # test the extra_metadata
    if expected['geometry'] == 'llc':
        emeta = {'has_faces': True, 'ny': 13*90, 'nx': 90,
                 'ny_facets': [3*90, 3*90, 90, 3*90, 3*90],
                 'face_facets': [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
                 'facet_orders': ['C', 'C', 'C', 'F', 'F'],
                 'face_offsets': [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
                 'transpose_face': [False, False, False,
                                    False, False, False, False,
                                    True, True, True, True, True, True]}
    else:
        emeta = None
    res = read_mds(basename, chunks="2D", use_dask=False,
                   use_mmap=False, extra_metadata=emeta)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.ndarray)

    # make sure endianness works
    res = read_mds(basename, use_dask=False, use_mmap=False)
    testval = res[prefix].newbyteorder('<')[0, 0]
    res_endian = read_mds(basename, use_mmap=False,
                          endian='<', use_dask=False)
    val_endian = res_endian[prefix][0, 0]
    np.testing.assert_allclose(testval, val_endian)

    # try reading with iteration number
    prefix = 'T'
    basename = os.path.join(dirname, prefix)
    iternum = expected['test_iternum']
    res = read_mds(basename, iternum=iternum)
    assert prefix in res
    assert isinstance(res[prefix], dask.array.core.Array)

    # try some options
    res = read_mds(basename, iternum=iternum, use_dask=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.memmap)

    res = read_mds(basename, iternum=iternum, use_dask=False,
                   use_mmap=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.ndarray)

    res = read_mds(basename, iternum=iternum, chunks="2D")
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], dask.array.core.Array)

    res = read_mds(basename, iternum=iternum, chunks="2D",
                   use_dask=False)
    assert isinstance(res, dict)
    assert prefix in res
    print(type(res[prefix]))
    assert isinstance(res[prefix], np.ndarray)  # should be memmap

    res = read_mds(basename, iternum=iternum, chunks="2D",
                   use_dask=False, use_mmap=False)
    assert isinstance(res, dict)
    assert prefix in res
    assert isinstance(res[prefix], np.ndarray)

    # check fails with bad nx, ny
    if expected['geometry'] == 'llc':
        emeta.update({'ny': 13*270, 'nx': 270})
        with pytest.raises(AssertionError):
            res = read_mds(basename, iternum=iternum, chunks="2D",
                           use_dask=False,
                           use_mmap=False, extra_metadata=emeta)


def test_read_mds_tokens(mds_datadirs_with_diagnostics):
    from xmitgcm.utils import read_mds
    dirname, expected = mds_datadirs_with_diagnostics
    diagnostics = expected['diagnostics']
    prefix = diagnostics[0]
    basename = os.path.join(dirname, prefix)
    iternum = expected['test_iternum']
    data = read_mds(basename, iternum=iternum)
    dask_keys = set()
    for varname, da in data.items():
        keys = list(da.dask.keys())
        for k in keys:
            token = k[0]
            if 'mds' in token:
                dask_keys.add(token)
    assert len(dask_keys) == len(data)


def test_read_mds_no_meta(all_mds_datadirs):
    from xmitgcm.utils import read_mds
    dirname, expected = all_mds_datadirs
    shape = expected['shape']
    ny, nx = shape[-2:]
    if len(shape) == 4:
        # we have an llc
        nz, nface = shape[:2]
        ny = nx*nface
    else:
        nz = shape[0]
    dtype = expected['dtype']

    shape_2d = (ny, nx)
    shape_3d = shape_2d if nz == 1 else (nz,) + shape_2d

    prefixes = {'XC': shape_2d, 'hFacC': shape_3d}

    for prefix, shape in prefixes.items():
        basename = os.path.join(dirname, prefix)
        with hide_file(dirname, prefix + '.meta'):
            # can't read without specifying shape and dtype
            with pytest.raises(IOError):
                res = read_mds(basename)
            res = read_mds(basename, shape=shape, dtype=dtype, legacy=True)
            assert isinstance(res, dict)
            assert prefix in res
            # should be dask by default
            assert isinstance(res[prefix], dask.array.core.Array)
            assert res[prefix].shape == shape

            res = read_mds(basename, shape=shape, dtype=dtype, legacy=False)
            assert isinstance(res, dict)
            assert prefix in res
            # should be dask by default
            assert isinstance(res[prefix], dask.array.core.Array)
            assert res[prefix].shape == (1,) + shape


@pytest.mark.parametrize("method", ["smallchunks", "bigchunks"])
@pytest.mark.parametrize("memmap", [True, False])
def test_read_raw_data_llc(llc_mds_datadirs, method, memmap):
    if memmap and method == 'smallchunks':
        pytest.skip("Using `method='smallchunks` with `memmap=True` "
                    "opens too many files.")

    dirname, expected = llc_mds_datadirs

    from xmitgcm.utils import read_3d_llc_data

    shape = expected['shape']
    nz, nface, ny, nx = shape
    # the function will also return a nrecs dimension
    nrecs = 1
    shape = (nrecs,) + shape

    dtype = expected['dtype'].newbyteorder('>')

    # if we use memmap=True, we open too many files
    kwargs = dict(method=method, dtype=dtype, memmap=memmap)

    fname = os.path.join(dirname, 'T.%010d.data' % expected['test_iternum'])
    data = read_3d_llc_data(fname, nz, nx, **kwargs)
    assert data.shape == shape
    dc = data.compute()
    assert dc.shape == shape
    # once computed, all arrays are ndarray, even if backed by memmap
    assert isinstance(dc, np.ndarray)

    fname = os.path.join(dirname, 'XC.data')
    data = read_3d_llc_data(fname, 1, nx, **kwargs)
    # the z dimension is squeezed out by MDS,
    # so the function matches that behavior
    shape_2d = (shape[0],) + shape[2:]
    assert data.shape == shape_2d
    assert data.compute().shape == shape_2d


@pytest.mark.parametrize("memmap", [True, False])
def test_read_xyz_chunk(all_mds_datadirs, memmap):

    from xmitgcm.utils import _read_xyz_chunk

    dirname, expected = all_mds_datadirs

    file_metadata = expected
    file_metadata.update({'filename': dirname + '/' + 'T.' +
                          str(file_metadata['test_iternum']).zfill(10) +
                          '.data', 'vars': ['T'], 'endian': '>'})
    # set the size of dimensions (could be changed in _experiments)
    if file_metadata['geometry'] in ['llc']:
        file_metadata.update({'nx': file_metadata['shape'][3],
                              'ny': file_metadata['shape'][2],
                              'nface': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'has_faces': True})
        # function not designed for llc grids, except 1d variables
        with pytest.raises(ValueError):
            data = _read_xyz_chunk('T', file_metadata, use_mmap=memmap)
    else:
        file_metadata.update({'nx': file_metadata['shape'][2],
                              'ny': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'dims_vars': [('nz', 'ny', 'nx')],
                              'has_faces': False})

        data = _read_xyz_chunk('T', file_metadata, use_mmap=memmap)

        if memmap:
            assert isinstance(data, np.memmap)
        else:
            assert isinstance(data, np.ndarray)

        # test it fails for too large number of records
        with pytest.raises(ValueError):
            data = _read_xyz_chunk('T', file_metadata, rec=1, use_mmap=memmap)

    # test 1d variable
    file_metadata.update({'filename': dirname + '/' + 'RC' + '.data',
                          'vars': ['RC'], 'nx': 1, 'ny': 1,
                          'dims_vars': [('nz', 'ny', 'nx')]})

    data = _read_xyz_chunk('RC', file_metadata, use_mmap=memmap)
    if memmap:
        assert isinstance(data, np.memmap)
    else:
        assert isinstance(data, np.ndarray)


@pytest.mark.parametrize("memmap", [True, False])
def test_read_xy_chunk(all_mds_datadirs, memmap):

    from xmitgcm.utils import _read_xy_chunk

    dirname, expected = all_mds_datadirs

    file_metadata = expected
    file_metadata.update({'filename': dirname + '/' + 'T.' +
                          str(file_metadata['test_iternum']).zfill(10) +
                          '.data', 'vars': ['T'], 'endian': '>'})
    # set the size of dimensions (could be changed in _experiments)
    if file_metadata['geometry'] in ['llc']:
        nx = file_metadata['shape'][3]
        file_metadata.update({'nx': file_metadata['shape'][3],
                              'ny': file_metadata['shape'][2],
                              'nface': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'dims_vars': [('nz', 'nface', 'ny', 'nx')],
                              'has_faces': True,
                              'ny_facets': [3*nx, 3*nx, nx, 3*nx, 3*nx],
                              'face_facets':
                              [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
                              'facet_orders': ['C', 'C', 'C', 'F', 'F'],
                              'face_offsets':
                              [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
                              'transpose_face': [False, False, False, False,
                                                 False, False, False, True,
                                                 True, True, True, True,
                                                 True]})
    else:
        file_metadata.update({'nx': file_metadata['shape'][2],
                              'ny': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'dims_vars': [('nz', 'ny', 'nx')],
                              'has_faces': False})

    data = _read_xy_chunk('T', file_metadata, use_mmap=memmap)

    if memmap:
        assert isinstance(data, np.memmap)
    else:
        assert isinstance(data, np.ndarray)

    # test it fails for too large number of records
    with pytest.raises(ValueError):
        data = _read_xy_chunk('T', file_metadata, rec=1, use_mmap=memmap)
    # test it fails for too large number of levels
    with pytest.raises(ValueError):
        data = _read_xy_chunk('T', file_metadata, lev=9999, use_mmap=memmap)

    # those tests are only available for llc experiment:
    # test reading in multi-variable files
    if expected['geometry'] not in ['llc']:
        pass
    else:
        dimsvar = []
        for kk in np.arange(25):
            dimsvar.append(('ny', 'nx'))
        file_metadata.update({'filename': dirname + '/' +
                              'state_2d_set1.0000000008.data',
                              'vars': expected['diagnostics'][1],
                              'dims_vars': dimsvar})

        for kface in np.arange(13):
            data = _read_xy_chunk('MXLDEPTH', file_metadata, face=kface,
                                  use_mmap=memmap)

        if memmap:
            assert isinstance(data, np.memmap)
        else:
            assert isinstance(data, np.ndarray)


@pytest.mark.parametrize("memmap", [True, False])
@pytest.mark.parametrize("usedask", [True, False])
def test_read_2D_chunks(all_mds_datadirs, memmap, usedask):

    from xmitgcm.utils import read_2D_chunks

    dirname, expected = all_mds_datadirs

    file_metadata = expected
    file_metadata.update({'filename': dirname + '/' + 'T.' +
                          str(file_metadata['test_iternum']).zfill(10) +
                          '.data', 'vars': ['T'], 'endian': '>'})
    # set the size of dimensions (could be changed in _experiments)
    if file_metadata['geometry'] in ['llc']:
        nx = file_metadata['shape'][3]
        file_metadata.update({'nx': file_metadata['shape'][3],
                              'ny': file_metadata['shape'][2],
                              'nface': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'nt': 1,
                              'dims_vars': [('nz', 'nface', 'ny', 'nx')],
                              'has_faces': True,
                              'ny_facets': [3*nx, 3*nx, nx, 3*nx, 3*nx],
                              'face_facets':
                              [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
                              'facet_orders': ['C', 'C', 'C', 'F', 'F'],
                              'face_offsets':
                              [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
                              'transpose_face': [False, False, False, False,
                                                 False, False, False, True,
                                                 True, True, True, True,
                                                 True]})
    else:
        file_metadata.update({'nx': file_metadata['shape'][2],
                              'ny': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'nt': 1,
                              'dims_vars': [('nz', 'ny', 'nx')],
                              'has_faces': False})

    data = read_2D_chunks('T', file_metadata, use_mmap=memmap,
                          use_dask=usedask)
    if usedask:
        assert isinstance(data, dask.array.core.Array)
        data.compute()  # check accessing data works
    else:
        if memmap:
            assert isinstance(data, np.ndarray)  # should be memmap, need fix
        else:
            assert isinstance(data, np.ndarray)

    # test 1d variable
    file_metadata.update({'filename': dirname + '/' + 'RC' + '.data',
                          'vars': ['RC'], 'nx': 1, 'ny': 1})

    data = read_2D_chunks('RC', file_metadata, use_mmap=memmap,
                          use_dask=usedask)
    if usedask:
        assert isinstance(data, dask.array.core.Array)
        data.compute()
    else:
        if memmap:
            assert isinstance(data, np.ndarray)  # reshape triggers evaluation
        else:
            assert isinstance(data, np.ndarray)


@pytest.mark.parametrize("memmap", [True, False])
@pytest.mark.parametrize("usedask", [True, False])
def test_read_3D_chunks(all_mds_datadirs, memmap, usedask):

    from xmitgcm.utils import read_3D_chunks

    dirname, expected = all_mds_datadirs

    file_metadata = expected
    file_metadata.update({'filename': dirname + '/' + 'T.' +
                          str(file_metadata['test_iternum']).zfill(10) +
                          '.data', 'vars': ['T'], 'endian': '>'})
    # set the size of dimensions (could be changed in _experiments)
    if file_metadata['geometry'] in ['llc']:
        nx = file_metadata['shape'][3]
        file_metadata.update({'nx': file_metadata['shape'][3],
                              'ny': file_metadata['shape'][2],
                              'nface': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'nt': 1,
                              'dims_vars': [('nz', 'nface', 'ny', 'nx')],
                              'has_faces': True,
                              'ny_facets': [3*nx, 3*nx, nx, 3*nx, 3*nx],
                              'face_facets':
                              [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
                              'facet_orders': ['C', 'C', 'C', 'F', 'F'],
                              'face_offsets':
                              [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
                              'transpose_face': [False, False, False, False,
                                                 False, False, False, True,
                                                 True, True, True, True,
                                                 True]})
    else:
        file_metadata.update({'nx': file_metadata['shape'][2],
                              'ny': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'nt': 1,
                              'dims_vars': [('nz', 'ny', 'nx')],
                              'has_faces': False})

    if file_metadata['geometry'] in ['llc']:
        with pytest.raises(ValueError):
            data = read_3D_chunks('T', file_metadata, use_mmap=memmap,
                                  use_dask=usedask)
            if usedask:
                data.compute()
    else:
        data = read_3D_chunks('T', file_metadata, use_mmap=memmap,
                              use_dask=usedask)
        if usedask:
            assert isinstance(data, dask.array.core.Array)
            data.compute()
        else:
            if memmap:
                assert isinstance(data, np.memmap)
            else:
                assert isinstance(data, np.ndarray)

    # test 1d variable
    file_metadata.update({'filename': dirname + '/' + 'RC' + '.data',
                          'vars': ['RC'], 'nx': 1, 'ny': 1})

    data = read_3D_chunks('RC', file_metadata, use_mmap=memmap,
                          use_dask=usedask)
    if usedask:
        assert isinstance(data, dask.array.core.Array)
        data.compute()
    else:
        if memmap:
            assert isinstance(data, np.memmap)
        else:
            assert isinstance(data, np.ndarray)


@pytest.mark.parametrize("memmap", [True, False])
@pytest.mark.parametrize("usedask", [True, False])
def test_read_all_variables(all_mds_datadirs, memmap, usedask):

    from xmitgcm.utils import read_all_variables

    dirname, expected = all_mds_datadirs

    file_metadata = expected
    # test single variable in file
    file_metadata.update({'filename': dirname + '/' + 'T.' +
                          str(file_metadata['test_iternum']).zfill(10) +
                          '.data', 'vars': ['T'], 'endian': '>'})
    # set the size of dimensions (could be changed in _experiments)
    if file_metadata['geometry'] in ['llc']:
        nx = file_metadata['shape'][3]
        file_metadata.update({'nx': file_metadata['shape'][3],
                              'ny': file_metadata['shape'][2],
                              'nface': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'nt': 1,
                              'dims_vars': [('nz', 'nface', 'ny', 'nx')],
                              'has_faces': True,
                              'ny_facets': [3*nx, 3*nx, nx, 3*nx, 3*nx],
                              'face_facets':
                              [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
                              'facet_orders': ['C', 'C', 'C', 'F', 'F'],
                              'face_offsets':
                              [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
                              'transpose_face': [False, False, False, False,
                                                 False, False, False, True,
                                                 True, True, True, True,
                                                 True]})
    else:
        file_metadata.update({'nx': file_metadata['shape'][2],
                              'ny': file_metadata['shape'][1],
                              'nz': file_metadata['shape'][0],
                              'nt': 1,
                              'dims_vars': [('nz', 'ny', 'nx')],
                              'has_faces': False})

    # test 3D chunks, fails on llc but not others
    if file_metadata['geometry'] in ['llc']:
        with pytest.raises(ValueError):
            dataset = read_all_variables(file_metadata['vars'], file_metadata,
                                         use_mmap=memmap, use_dask=usedask,
                                         chunks="3D")
            if usedask:
                dataset[0].compute()
    else:
        dataset = read_all_variables(file_metadata['vars'], file_metadata,
                                     use_mmap=memmap, use_dask=usedask,
                                     chunks="3D")

        assert isinstance(dataset, list)
        assert len(dataset) == len(file_metadata['vars'])
        if usedask:
            assert isinstance(dataset[0], dask.array.core.Array)
        else:
            if memmap:
                assert isinstance(dataset[0], np.memmap)
            else:
                assert isinstance(dataset[0], np.ndarray)

    # test 2D chunks
    dataset = read_all_variables(file_metadata['vars'], file_metadata,
                                 use_mmap=memmap, use_dask=usedask,
                                 chunks="2D")

    assert isinstance(dataset, list)
    assert len(dataset) == len(file_metadata['vars'])
    if usedask:
        assert isinstance(dataset[0], dask.array.core.Array)
    else:
        if memmap:
            # should be memmap, needs fix
            assert isinstance(dataset[0], np.ndarray)
        else:
            assert isinstance(dataset[0], np.ndarray)

    # test multiple variables in file
    # those tests are only available for llc experiment:
    # test reading in multi-variable files
    if expected['geometry'] not in ['llc']:
        pass
    else:
        dimsvar = []
        for kk in np.arange(25):
            dimsvar.append(('ny', 'nx'))
        file_metadata.update({'filename': dirname + '/' +
                              'state_2d_set1.0000000008.data',
                              'vars': expected['diagnostics'][1],
                              'dims_vars': dimsvar})

    dataset = read_all_variables(file_metadata['vars'], file_metadata,
                                 use_mmap=memmap, use_dask=usedask,
                                 chunks="2D")

    assert isinstance(dataset, list)
    assert len(dataset) == len(file_metadata['vars'])
    if usedask:
        assert isinstance(dataset[0], dask.array.core.Array)
    else:
        if memmap:
            # should be memmap, needs fix
            assert isinstance(dataset[0], np.ndarray)
        else:
            assert isinstance(dataset[0], np.ndarray)


@pytest.mark.parametrize("dtype", ['>d', '>f', '>i'])
@pytest.mark.parametrize("memmap", [True, False])
def test_pad_array(tmpdir, memmap, dtype):

    from xmitgcm.utils import _pad_array
    import struct

    # create test data
    gendata = np.array([[1, 2], [3, 4]])

    # write to a file
    datafile = tmpdir.join("testdata")
    fname = str(datafile)
    fid = open(fname, "wb")
    flatdata = gendata.flatten()
    for kk in np.arange(len(flatdata)):
        tmp = struct.pack(dtype, flatdata[kk])
        fid.write(tmp)
    fid.close()

    # then read it
    if memmap:
        data = np.memmap(fname, dtype=dtype, mode='r',
                         shape=(2, 2,), order='C')
    else:
        data = np.fromfile(fname, dtype=dtype)
        data = data.reshape((2, 2,))

    # check my original data
    ny, nx = data.shape
    assert data.shape == (ny, nx)
    assert data.min() == 1
    assert data.max() == 4

    # test no padding
    file_metadata = {}
    data_padded = _pad_array(data, file_metadata)
    assert data_padded.shape == (2, 2)
    if memmap:
        assert isinstance(data_padded, np.memmap)
    else:
        assert isinstance(data_padded, np.ndarray)

    # test padding before
    file_metadata = {'pad_before_y': 2, 'has_faces': False, 'nx': nx}
    data_padded = _pad_array(data, file_metadata)
    assert isinstance(data_padded, np.ndarray)
    assert data_padded.shape == (4, 2)
    assert data_padded.min() == 0
    assert data_padded.max() == 4
    assert data_padded[2, 0] == 1
    assert data_padded[3, 1] == 4

    file_metadata = {'pad_before_y': [2, 3], 'has_faces': True, 'nx': nx,
                     'face_facets': [0, 1]}
    data_padded = _pad_array(data, file_metadata, face=0)
    assert isinstance(data_padded, np.ndarray)
    assert data_padded.shape == (4, 2)
    assert data_padded.min() == 0
    assert data_padded.max() == 4
    assert data_padded[2, 0] == 1
    assert data_padded[3, 1] == 4

    data_padded = _pad_array(data, file_metadata, face=1)
    assert isinstance(data_padded, np.ndarray)
    assert data_padded.shape == (5, 2)
    assert data_padded.min() == 0
    assert data_padded.max() == 4
    assert data_padded[3, 0] == 1
    assert data_padded[4, 1] == 4

    # test padding after
    file_metadata = {'pad_after_y': 2, 'has_faces': False, 'nx': nx}
    data_padded = _pad_array(data, file_metadata)
    assert isinstance(data_padded, np.ndarray)
    assert data_padded.shape == (4, 2)
    assert data_padded.min() == 0
    assert data_padded.max() == 4
    assert data_padded[0, 0] == 1
    assert data_padded[1, 1] == 4

    file_metadata = {'pad_after_y': [2, 3], 'has_faces': True, 'nx': nx,
                     'face_facets': [0, 1]}
    data_padded = _pad_array(data, file_metadata, face=0)
    assert isinstance(data_padded, np.ndarray)
    assert data_padded.shape == (4, 2)
    assert data_padded.min() == 0
    assert data_padded.max() == 4
    assert data_padded[0, 0] == 1
    assert data_padded[1, 1] == 4

    data_padded = _pad_array(data, file_metadata, face=1)
    assert isinstance(data_padded, np.ndarray)
    assert data_padded.shape == (5, 2)
    assert data_padded.min() == 0
    assert data_padded.max() == 4
    assert data_padded[0, 0] == 1
    assert data_padded[1, 1] == 4


def test_parse_diagnostics(all_mds_datadirs, layers_mds_datadirs):
    """Make sure we can parse the available_diagnostics.log file."""
    from xmitgcm.utils import parse_available_diagnostics
    dirname, expected = all_mds_datadirs
    diagnostics_fname = os.path.join(dirname, 'available_diagnostics.log')
    ad = parse_available_diagnostics(diagnostics_fname)

    # a somewhat random sampling of diagnostics
    expected_diags = {
        'UVEL': {'dims': ['k', 'j', 'i_g'],
                 'attrs': {'units': 'm/s',
                           'long_name': 'Zonal Component of Velocity (m/s)',
                           'standard_name': 'UVEL',
                           'mate': 'VVEL'}},
        'TFLUX': {'dims': ['j', 'i'],
                  'attrs': {'units': 'W/m^2',
                            'long_name': 'total heat flux (match heat-content '
                            'variations), >0 increases theta',
                            'standard_name': 'TFLUX'}}
    }

    for key, val in expected_diags.items():
        assert ad[key] == val

    # test layers
    dirname, expected = layers_mds_datadirs
    diagnostics_fname = os.path.join(dirname, 'available_diagnostics.log')
    ad = parse_available_diagnostics(diagnostics_fname)

    expected_diags = {
        'LaUH1RHO': {'dims': ['_UNKNOWN_', 'j', 'i_g'],
                     'attrs': {'units': 'm.m/s',
                               'long_name': 'Layer Integrated  zonal Transport (UH, m^2/s)',
                               'standard_name': 'LaUH1RHO',
                               'mate': 'LaVH1RHO'}},
    }

    for key, val in expected_diags.items():
        assert ad[key] == val


@pytest.mark.parametrize("domain", ['llc', 'aste', 'cs'])
@pytest.mark.parametrize("nx", [90, 270])
def test_get_extra_metadata(domain, nx):
    from xmitgcm.utils import get_extra_metadata
    em = get_extra_metadata(domain=domain, nx=nx)
    assert type(em) == dict

    with pytest.raises(ValueError):
        em = get_extra_metadata(domain='notinlist', nx=nx)


@pytest.mark.parametrize("usedask", [True, False])
def test_get_grid_from_input(all_grid_datadirs, usedask):
    from xmitgcm.utils import get_grid_from_input, get_extra_metadata
    from xmitgcm.utils import read_raw_data
    dirname, expected = all_grid_datadirs
    md = get_extra_metadata(domain=expected['domain'], nx=expected['nx'])
    ds = get_grid_from_input(dirname + '/' + expected['gridfile'],
                             geometry=expected['geometry'],
                             dtype=np.dtype('d'), endian='>',
                             use_dask=usedask,
                             extra_metadata=md)
    # test types
    assert type(ds) == xarray.Dataset
    assert type(ds['XC']) == xarray.core.dataarray.DataArray

    if usedask:
        ds.load()

    # check all variables are in
    expected_variables = ['XC', 'YC', 'DXF', 'DYF', 'RAC',
                          'XG', 'YG', 'DXV', 'DYU', 'RAZ',
                          'DXC', 'DYC', 'RAW', 'RAS', 'DXG', 'DYG']

    for var in expected_variables:
        assert type(ds[var]) == xarray.core.dataarray.DataArray
        assert ds[var].values.shape == expected['shape']

    # check we don't leave points behind
    if expected['geometry'] == 'llc':
        nx = expected['nx'] + 1
        nvars = len(expected_variables)
        sizeofd = 8

        grid = expected['gridfile']
        grid1 = dirname + '/' + grid.replace('<NFACET>', '001')
        grid2 = dirname + '/' + grid.replace('<NFACET>', '002')
        grid3 = dirname + '/' + grid.replace('<NFACET>', '003')
        grid4 = dirname + '/' + grid.replace('<NFACET>', '004')
        grid5 = dirname + '/' + grid.replace('<NFACET>', '005')

        size1 = os.path.getsize(grid1)
        size2 = os.path.getsize(grid2)
        size3 = os.path.getsize(grid3)
        size4 = os.path.getsize(grid4)
        size5 = os.path.getsize(grid5)

        ny1 = int(size1 / sizeofd / nvars / nx)
        ny2 = int(size2 / sizeofd / nvars / nx)
        ny3 = int(size3 / sizeofd / nvars / nx)
        ny4 = int(size4 / sizeofd / nvars / nx)
        ny5 = int(size5 / sizeofd / nvars / nx)

        xc1 = read_raw_data(grid1, dtype=np.dtype('>d'), shape=(ny1, nx),
                            partial_read=True)
        xc2 = read_raw_data(grid2, dtype=np.dtype('>d'), shape=(ny2, nx),
                            partial_read=True)
        xc3 = read_raw_data(grid3, dtype=np.dtype('>d'), shape=(ny3, nx),
                            partial_read=True)
        xc4 = read_raw_data(grid4, dtype=np.dtype('>d'), shape=(ny4, nx),
                            order='F', partial_read=True)
        xc5 = read_raw_data(grid5, dtype=np.dtype('>d'), shape=(ny5, nx),
                            order='F', partial_read=True)

        yc1 = read_raw_data(grid1, dtype=np.dtype('>d'), shape=(ny1, nx),
                            partial_read=True, offset=nx*ny1*sizeofd)
        yc2 = read_raw_data(grid2, dtype=np.dtype('>d'), shape=(ny2, nx),
                            partial_read=True, offset=nx*ny2*sizeofd)
        yc3 = read_raw_data(grid3, dtype=np.dtype('>d'), shape=(ny3, nx),
                            partial_read=True, offset=nx*ny3*sizeofd)
        yc4 = read_raw_data(grid4, dtype=np.dtype('>d'), shape=(ny4, nx),
                            order='F', partial_read=True,
                            offset=nx*ny4*sizeofd)
        yc5 = read_raw_data(grid5, dtype=np.dtype('>d'), shape=(ny5, nx),
                            order='F', partial_read=True,
                            offset=nx*ny5*sizeofd)

        xc = np.concatenate([xc1[:-1, :-1].flatten(), xc2[:-1, :-1].flatten(),
                             xc3[:-1, :-1].flatten(), xc4[:-1, :-1].flatten(),
                             xc5[:-1, :-1].flatten()])

        yc = np.concatenate([yc1[:-1, :-1].flatten(), yc2[:-1, :-1].flatten(),
                             yc3[:-1, :-1].flatten(), yc4[:-1, :-1].flatten(),
                             yc5[:-1, :-1].flatten()])

        xc_from_ds = ds['XC'].values.flatten()
        yc_from_ds = ds['YC'].values.flatten()

        assert xc.min() == xc_from_ds.min()
        assert xc.max() == xc_from_ds.max()
        assert yc.min() == yc_from_ds.min()
        assert yc.max() == yc_from_ds.max()

    # passing llc without metadata should fail
    if expected['geometry'] == 'llc':
        with pytest.raises(ValueError):
            ds = get_grid_from_input(dirname + '/' + expected['gridfile'],
                                     geometry=expected['geometry'],
                                     dtype=np.dtype('d'), endian='>',
                                     use_dask=False,
                                     extra_metadata=None)


@pytest.mark.parametrize("dtype", [np.dtype('d'), np.dtype('f')])
def test_write_to_binary(dtype):
    from xmitgcm.utils import write_to_binary
    import sys

    data = np.arange(2)
    # write
    write_to_binary(data, 'tmp.bin', dtype=dtype)
    # read
    if dtype == np.dtype('f'):
        tmp = np.fromfile('tmp.bin', '>f')
    elif dtype == np.dtype('d'):
        tmp = np.fromfile('tmp.bin', '>d')
    # check
    assert len(data) == len(tmp)
    assert data[0] == tmp[0]
    assert data[1] == tmp[1]
    os.remove('tmp.bin')


@pytest.mark.parametrize("possible_concat_dims", [['i', 'i_g'], ['j', 'j_g']])
def test_find_concat_dim(possible_concat_dims):
    from xmitgcm.utils import find_concat_dim

    # this array contains a concat dim
    a = xarray.DataArray(np.empty((2, 3, 4)), dims=['k', 'j', 'i'])
    out = find_concat_dim(a, possible_concat_dims)
    assert out in possible_concat_dims

    b = xarray.DataArray(np.empty((2, 3, 4)), dims=['k', 'g', 'b'])
    out = find_concat_dim(b, possible_concat_dims)
    assert out is None


@pytest.mark.parametrize("domain", ['aste', 'llc'])
@pytest.mark.parametrize("nx", [90, 270])
def test_find_concat_dim_facet(domain, nx):
    from xmitgcm.utils import find_concat_dim_facet, get_extra_metadata
    md = get_extra_metadata(domain=domain, nx=nx)
    nfacets = len(md['ny_facets'])

    for facet in range(5):
        da = xarray.DataArray(np.empty((nfacets, md['ny_facets'][facet], nx)),
                              dims=['face', 'j', 'i'])
        concat_dim, non_concat_dim = find_concat_dim_facet(da, facet, md)

        print(concat_dim, non_concat_dim)
        if md['facet_orders'][facet] == 'C':
            assert concat_dim == 'j'
            assert non_concat_dim == 'i'
        elif md['facet_orders'][facet] == 'F':
            assert concat_dim == 'i'
            assert non_concat_dim == 'j'


@pytest.mark.parametrize("domain", ['aste', 'llc'])
@pytest.mark.parametrize("nx", [90, 270])
def test_rebuild_llc_facets(domain, nx):
    from xmitgcm.utils import rebuild_llc_facets, get_extra_metadata

    md = get_extra_metadata(domain=domain, nx=nx)
    nfaces = len(md['transpose_face'])

    da = xarray.DataArray(np.empty((nfaces, nx, nx)),
                          dims=['face', 'j', 'i'])

    facets = rebuild_llc_facets(da, md)

    for facet in range(5):
        # test we get the original size
        if md['facet_orders'][facet] == 'C':
            expected_shape = (md['ny_facets'][facet], nx,)
        elif md['facet_orders'][facet] == 'F':
            expected_shape = (nx, md['ny_facets'][facet], )
        if domain == 'aste' and facet == 1:  # this facet is empty
            pass
        else:
            assert facets['facet' + str(facet)].shape == expected_shape


def test_llc_facets_2d_to_compact(llc_mds_datadirs):
    from xmitgcm.utils import llc_facets_2d_to_compact, get_extra_metadata
    from xmitgcm.utils import rebuild_llc_facets, read_raw_data
    from xmitgcm.utils import write_to_binary
    from xmitgcm import open_mdsdataset

    dirname, expected = llc_mds_datadirs

    # open dataset
    ds = open_mdsdataset(dirname,
                         iters=expected['test_iternum'],
                         geometry=expected['geometry'])

    nt, nfaces, ny, nx = expected['shape']
    md = get_extra_metadata(domain=expected['geometry'], nx=nx)
    # split in facets
    facets = rebuild_llc_facets(ds['XC'], md)
    flatdata = llc_facets_2d_to_compact(facets, md)
    # compare with raw data
    raw = read_raw_data(dirname + '/XC.data', np.dtype('>f'), (nfaces, ny, nx))
    flatraw = raw.flatten()

    assert len(flatdata) == len(flatraw)
    assert flatdata.min() == flatraw.min()
    assert flatdata.max() == flatraw.max()

    # write new file
    write_to_binary(flatdata, 'tmp.bin', dtype=np.dtype('f'))
    md5new = file_md5_checksum('tmp.bin')
    md5old = file_md5_checksum(dirname + '/XC.data')
    assert md5new == md5old
    os.remove('tmp.bin')


def test_llc_facets_3d_spatial_to_compact(llc_mds_datadirs):
    from xmitgcm.utils import llc_facets_3d_spatial_to_compact
    from xmitgcm.utils import get_extra_metadata
    from xmitgcm.utils import rebuild_llc_facets, read_raw_data
    from xmitgcm.utils import write_to_binary
    from xmitgcm import open_mdsdataset

    dirname, expected = llc_mds_datadirs

    # open dataset
    ds = open_mdsdataset(dirname,
                         iters=expected['test_iternum'],
                         geometry=expected['geometry'])

    nz, nfaces, ny, nx = expected['shape']
    md = get_extra_metadata(domain=expected['geometry'], nx=nx)
    # split in facets
    facets = rebuild_llc_facets(ds['T'], md)
    flatdata = llc_facets_3d_spatial_to_compact(facets, 'k', md)
    # compare with raw data
    raw = read_raw_data(dirname + '/T.' +
                        str(expected['test_iternum']).zfill(10) + '.data',
                        np.dtype('>f'), (nz, nfaces, ny, nx))
    flatraw = raw.flatten()

    assert len(flatdata) == len(flatraw)
    assert flatdata.min() == flatraw.min()
    assert flatdata.max() == flatraw.max()

    # write new file
    write_to_binary(flatdata, 'tmp.bin', dtype=np.dtype('f'))
    md5new = file_md5_checksum('tmp.bin')
    md5old = file_md5_checksum(dirname + '/T.' +
                               str(expected['test_iternum']).zfill(10) +
                               '.data')
    assert md5new == md5old
    os.remove('tmp.bin')


def test_parse_namelist(tmpdir, mds_datadirs_with_inputfiles):
    from xmitgcm.utils import parse_namelist

    dirname, expected = mds_datadirs_with_inputfiles
    exp_vals = expected['expected_namelistvals']
    # read namelist "data"
    data = parse_namelist(os.path.join(dirname, 'data'))

    assert data['PARM01']['eosType'] == exp_vals['eosType']
    assert data['PARM01']['viscAh'] == exp_vals['viscAh']
    assert data['PARM03']['niter0'] == exp_vals['niter0']
    assert data['PARM04']['delX'] == exp_vals['delX']

    diags = parse_namelist(os.path.join(dirname, 'data.diagnostics'))

    assert diags['DIAGNOSTICS_LIST']['levels'] == exp_vals['levels']
    assert diags['DIAGNOSTICS_LIST']['fileName'] == exp_vals['fileName']

    pkgs = parse_namelist(os.path.join(dirname, 'data.pkg'))

    assert pkgs['PACKAGES']['useDiagnostics'] is exp_vals['useDiagnostics']

    with open(os.path.join(str(tmpdir), 'invalid_namelists'), 'w') as f:
        f.write("# This is an invalid namelist\n"
                " &PARM01\n"
                " tRef= 12*10.,\n"
                " sRef= 15*1e3,\n"
                " cosPower=invalid\n"
                " viscAr=1.E-3,\n"
                " &\n")

    with pytest.warns(UserWarning, match='Unable to read value'):
        data = parse_namelist(os.path.join(str(tmpdir), 'invalid_namelists'),
                              silence_errors=True)
    assert data['PARM01']['cosPower'] is None


    with pytest.raises(ValueError, match='Unable to read value'):
        parse_namelist(os.path.join(str(tmpdir), 'invalid_namelists'),
                       silence_errors=False)



